from __future__ import print_function, division
import numpy
from numpy import pi, abs, sqrt, cos, sin, arccos, arctan, tan, cbrt
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo
from PyAstronomy import constants as _PC
import copy


class MarkleyKESolver:
    """
    Solve Kepler's Equation.

    This class implements a solver for Kepler's Equation:
    .. math::
        M = E - sin(E),
    where M is the "mean anomaly" and E is the "eccentric anomaly".
    The implementation follows the prescription given by
    Markley (Markley 1995, CeMDA, 63, 101).
    """

    pi2 = pi**2

    def _alpha(self, e, M):
        """
        Solve Eq. 20
        """
        return (3. * self.pi2 + 1.6 * pi * (pi - abs(M)) / (1. + e)) / (self.pi2 - 6.)

    def _d(self, alpha, e):
        """
        Solve Eq. 5
        """
        return 3. * (1. - e) + alpha * e

    def _r(self, alpha, d, M, e):
        """
        Solve Eq. 10
        """
        return 3. * alpha * d * (d - 1. + e) * M + M**3

    def _q(self, alpha, d, e, M):
        """
        Solve Eq. 9
        """
        return 2. * alpha * d * (1. - e) - M**2

    def _w(self, r, q):
        """
        Solve Eq. 14
        """
        return (abs(r) + sqrt(q**3 + r**2))**(2. / 3.)

    def _E1(self, d, r, w, q, M):
        """
        Solve Eq. 15
        """
        return (2. * r * w / (w**2 + w * q + q**2) + M) / d

    def _f01234(self, e, E, M):
        """
        Solve Eq. 21, 25, 26, 27, and 28 (f, f', f'', f''', and f'''')
        """
        f0 = E - e * sin(E) - M
        f1 = 1. - e * cos(E)
        f2 = e * sin(E)
        return f0, f1, f2, 1. - f1, -f2

    def _d3(self, E, f):
        """
        Solve Eq. 22 
        """
        return -f[0] / (f[1] - 0.5 * f[0] * f[2] / f[1])

    def _d4(self, E, f, d3):
        """
        Solve Eq. 23
        """
        return -f[0] / (f[1] + 0.5 * d3 * f[2] + (d3**2) * f[3] / 6.)

    def _d5(self, E, f, d4):
        """
        Solve Eq. 24
        """
        return -f[0] / (f[1] + 0.5 * d4 * f[2] + d4**2 * f[3] / 6. + d4**3 * f[4] / 24.)

    def precisionTest(self, trials=10000):
        """
        Carry out a test of the achieved precision.

        Generate random numbers for the mean anomaly
        and the eccentricity and calculate the eccentric
        anomaly. Use the number in Kepler's Equation
        and compare the resulting mean anomaly with
        the input.

        Parameters
        ----------
        trials : int, optional
            The number of trial evaluations.

        Returns
        -------
        Deviation : float
            The largest determined deviation between
            input and output mean anomaly.
        """
        maxdev = 0.0
        for i in smo.range(trials):
            M = numpy.random.random() * 100. - 50.
            e = numpy.random.random()
            E = self.getE(M, e)
            M1 = E - e * sin(E)
            M = M - numpy.floor(M / (2. * pi)) * 2 * pi
            maxdev = max(maxdev, abs(M - M1))
        print("Maximal deviation in resulting mean anomaly (M): ", maxdev)
        return maxdev

    def getE(self, M, e):
        """
        Solve Kepler's Equation for the "eccentric anomaly", E.

        Parameters
        ----------
        M : float
            Mean anomaly.
        e : float
            Eccentricity

        Returns
        -------
        Eccentric anomaly: float
            The solution of Kepler's Equation
        """
        # For the mean anomaly, use values between
        # -pi and pi.
        flip = False
        self.M = M - (numpy.floor(M / (2. * pi)) * 2. * pi)
        if self.M > pi:
            self.M = 2. * pi - self.M
            # Flip the sign of result
            # if this happened
            flip = True
        self.e = e
        if M == 0.0:
            return 0.0
        self.alpha = self._alpha(e, self.M)
        self.d = self._d(self.alpha, e)
        self.r = self._r(self.alpha, self.d, self.M, e)
        self.q = self._q(self.alpha, self.d, e, self.M)
        self.w = self._w(self.r, self.q)
        self.E1 = self._E1(self.d, self.r, self.w, self.q, self.M)
        self.f = self._f01234(e, self.E1, self.M)
        self.d3 = self._d3(self.E1, self.f)
        self.d4 = self._d4(self.E1, self.f, self.d3)
        self.d5 = self._d5(self.E1, self.f, self.d4)
        # Eq. 29
        self.E5 = self.E1 + self.d5
        if flip:
            self.E5 = 2. * pi - self.E5
        self.E = self.E5
        return self.E5


class KeplerEllipse(object):
    """
    Calculate a Kepler orbit.

    The definitions and most of the formulae used in this class
    derive from the book "Orbital Motion" by A.E. Roy.

    :Orientation of the ellipse in the coordinate system:
        One of the foci of the ellipse is in the origin of the coordinate system.
        For zero inclination the ellipse is located in the x-y plane.
        If the eccentricity is increased, the periastron will lie
        in +x direction. If the inclination is increased, the ellipse
        will be rotating around the x-axis, so that +y is rotated
        toward +z. An increase in Omega corresponds to a rotation
        around the z-axis so that +x is rotated toward +y.
        Changing `w`, i.e., the argument of the periastron, will
        not change the plane of the orbit, but rather represent a
        rotation of the orbit in the plane. In particular, the
        periapsis is shifted in the direction of motion.

    :Orbital angular momentum:
        For all parameters but semi-major axis and orbital period set to zero,
        the (orbital) angular momentum points into the +z direction. For an
        inclination of 90 deg (the remaining parameters remaining zero),
        it points in the -y direction.

    :Orientation of the ellipse in the sky:
        To project the ellipse onto the sky, the coordinate system
        should be oriented so that the +x direction points North and
        the +y direction points East (direction of increasing right
        ascension). The +z axis must be chosen so that the coordinate
        system becomes right handed. If the line of sight (LOS) points
        in the +z direction, i.e., the observer is located on the
        negative z axis, the parameters assume their conventional
        meaning.

    :The ascending and descending nodes:
        For systems outside the Solar System, the ascending node is the
        point where the body "crosses" the plane of the sky away from the
        observer. Likewise, the descending node is the point where the
        plane is crossed with the body approaching the observer. For the
        coordinate system described above and a value of zero for the longitude
        of the ascending node, the latter is in the North and rotates
        toward East (i.e., +y) when the longitude of the ascending node
        is increased.  

    :The argument and longitude of periapsis:
        The argument of periapsis is the angle between the ascending node
        and the periapsis of the body measured in the direction of motion.
        For exoplanets with circular orbits, for which no well-defined periapsis
        exists, the argument of periapsis is often chosen so that time
        of periapsis and central transit time coincide. For the planet, this
        is the case if the argument of periapsis is -90 deg. However, in the exoplanet
        literature, the argument of periapsis often refers to the *stellar* orbit
        (see, e.g., Pollacco et al. 2008, MNRAS 385, 1576-1584, Sect. 3.2.1). In
        this case, the corresponding value is +90 deg.
        The so-called longitude of the periapsis is given by the sum of the
        longitude of the ascending node and the argument of periapsis.
        Note that for example in the literature on eclipsing binaries the terms
        are not always clearly distinguished and what is referred to the longitude of
        periapsis may actually be the argument of periapsis in the above nomenclature.

    Parameters
    ----------
    a : float
        Semi-major axis. The units of this parameter define the
        length scale of the orbit. Common choices are AU, solar, or
        stellar radii.
    per : float
        Orbital period
    e : float, optional
        Orbital eccentricity (0-1).
    tau : float, optional
        Time of periapsis passage.
    Omega : float, optional
        Longitude of the ascending node [deg].
    w : float, optional
        Argument of periapsis [deg]. Note that the longitude
        if periapsis is given by Omega+w.
    i : float, optional
        Orbit inclination [deg].
    ks : Class object, optional
        The solver for Kepler's Equation. Default is the
        `MarkleyKESolver`. Each solver must have a `getE`
        method, which takes either a float or array of float
        of mean anomalies and the eccentricity, and returns
        the associated eccentric anomalies.

    Attributes
    ----------
    i : float
        Orbit inclination [deg].
    w : float
        Argument of periapsis [deg]
    Omega : float
        Longitude of the ascending node [deg]
    e : float
        Eccentricity
    a : float
        Semi-major axis
    per : float
        Orbital period
    tau : float
        Time of periapsis passage
    ks : Class object
        Solver for Kepler's equation
    _n : float
        Circular frequency.
    """

    def _getEccentricAnomaly(self, t):
        """
        Calculate eccentric anomaly.

        Parameters
        ----------
        t : array of float
            The times at which to calculate the eccentric anomaly, E.

        Returns
        -------
        E : Array of float
            The eccentric anomaly.
        """
        M = self.meanAnomaly(t)
        if not hasattr(t, "__iter__"):
            return self.ks.getE(M, self.e)
        E = numpy.zeros(len(t))
        for i in smo.range(len(t)):
            E[i] = self.ks.getE(M[i], self.e)
        return E

    def meanAnomaly(self, t):
        """
        Calculate the mean anomaly.

        Parameters
        ----------
        t : float or array
            The time axis.

        Returns
        -------
        Mean anomaly : float or array
            The mean anomaly (whether float or array
            depends on input).
        """
        return self._n * (t - self.tau)

    def radius(self, t, E=None):
        """
        Calculate the orbit radius.

        Parameters
        ----------
        t : float or array
            The time axis.
        E : float or array, optional
            If known, the eccentric anomaly corresponding
            to the time points. If not given, the numbers
            will be calculated.

        Returns
        -------
        Radius : float or array
            The orbit radius at the given points in time.
            Type depends on input type.
        """
        if E is None:
            E = self._getEccentricAnomaly(t)
        return self.a * (1. - self.e * cos(E))

    def trueAnomaly(self, t, E=None):
        """
        Calculate the true anomaly.

        Parameters
        ----------
        t : float or array
            The time axis.
        E : float or array, optional
            If known, the eccentric anomaly corresponding
            to the time points (note that `t` will be ignored then).
            If not given, the numbers will be calculated.

        Returns
        -------
        Radius : float or array
            The orbit radius at the given points in time.
            Type depends on input type.
        """
        if E is None:
            E = self._getEccentricAnomaly(t)
        f = arctan(sqrt((1. + self.e) / (1. - self.e)) * tan(E / 2.)) * 2.0
        return f
    
    def eccentricAnomaly(self, t):
        """
        Calculates the eccentric anomaly.

        Parameters
        ----------
        t : array of float
            The times at which to calculate the eccentric anomaly, E.

        Returns
        -------
        E : Array of float
            The eccentric anomaly.
        """
        return self._getEccentricAnomaly(t)
        
    def xyzPos(self, t, getTA=False):
        """
        Calculate orbit position.

        Parameters
        ----------
        t : float or array
            The time axis.
        getTA : boolean, optional
            If True, also returns the "true anomaly" as a function
            of time (default is False).

        Returns
        -------
        Position : array
            The x, y, and z coordinates of the body at the given time.
            If the input was an array, the output will be an array of
            shape (input-length, 3), holding the positions at the given
            times.
        True anomaly : float or array
            Is returned only if `getTA` is set to True. The true anomaly
            at the specified times.
        """
        E = self._getEccentricAnomaly(t)
        r = self.radius(t, E=E)
        f = self.trueAnomaly(None, E=E)
        wf = self._w + f
        cos_Omega = cos(self._Omega)
        sin_Omega = sin(self._Omega)
        cos_i = cos(self._i)
        sin_i = sin(self._i)
        if not hasattr(wf, "__iter__"):
            cos_wf = cos(wf)
            sin_wf = sin(wf)
            xyz = numpy.array([cos_Omega * cos_wf - sin_Omega * sin_wf * cos_i,
                               sin_Omega * cos_wf + cos_Omega * sin_wf * cos_i,
                               sin_wf * sin_i
                               ]) * r
        else:
            # Assume it is an array
            xyz = numpy.zeros((len(t), 3))
            for i in smo.range(len(t)):
                cos_wf = cos(wf[i])
                sin_wf = sin(wf[i])
                xyz[i, ::] = numpy.array([cos_Omega * cos_wf - sin_Omega * sin_wf * cos_i,
                                          sin_Omega * cos_wf + cos_Omega * sin_wf * cos_i,
                                          sin_wf * sin_i
                                          ]) * r[i]
        if not getTA:
            return xyz
        else:
            return xyz, f 

    def xyzVel(self, t):
        """
        Calculate orbit velocity.

        Parameters
        ----------
        t : float or array
            The time axis.

        Returns
        -------
        Velocity : array
            The x, y, and z components of the body's velocity at the
            given time. If the input was an array, the output will be
            an array of shape (input-length, 3), holding the velocity
            components at the given times. The unit is that of the
            semi-major axis divided by that of the period.
        """
        # From AE ROY "Orbital motion" p. 102
        cos_Omega = cos(self._Omega)
        sin_Omega = sin(self._Omega)
        cos_i = cos(self._i)
        sin_i = sin(self._i)
        cos_w = cos(self._w)
        sin_w = sin(self._w)

        E = self._getEccentricAnomaly(t)
        l1 = cos_Omega * cos_w - sin_Omega * sin_w * cos_i
        l2 = -cos_Omega * sin_w - sin_Omega * cos_w * cos_i
        m1 = sin_Omega * cos_w + cos_Omega * sin_w * cos_i
        m2 = -sin_Omega * sin_w + cos_Omega * cos_w * cos_i
        n1 = sin_w * sin_i
        n2 = cos_w * sin_i
        b = self.a * sqrt(1. - self.e**2)
        r = self.radius(t, E)
        nar = self._n * self.a / r
        if not hasattr(t, "__iter__"):
            bcos_E = b * cos(E)
            asin_E = self.a * sin(E)
            vel = nar * numpy.array([l2 * bcos_E - l1 * asin_E,
                                     m2 * bcos_E - m1 * asin_E,
                                     n2 * bcos_E - n1 * asin_E])
        else:
            # Assume it is an array
            vel = numpy.zeros((len(t), 3))
            for i in smo.range(len(t)):
                bcos_E = b * cos(E[i])
                asin_E = self.a * sin(E[i])
                vel[i, ::] = nar[i] * numpy.array([l2 * bcos_E - l1 * asin_E,
                                                   m2 * bcos_E - m1 * asin_E,
                                                   n2 * bcos_E - n1 * asin_E])
        return vel

    def xyzPeriastron(self):
        """
        The position of the periastron.

        Returns
        -------
        Periastron : array of float
            The x, y, and z coordinates of the periastron. 
        """
        return self.xyzPos(self.tau)

    def xyzApastron(self):
        """
        The position of the apastron.

        The apastron is the point of greatest distance.

        Returns
        -------
        Apastron : array of float
            The x, y, and z coordinates of the apastron
        """
        return self.xyzPos(self.tau + 0.5 * self.per)

    def xyzCenter(self):
        """
        Center of the ellipse

        Returns
        -------
        Center : array of float
            x, y, and z coordinates of the center of the Ellipse.
        """
        return (self.xyzPeriastron() + self.xyzApastron()) / 2.0

    def xyzFoci(self):
        """
        Calculate the foci of the ellipse

        Returns
        -------
        Foci : Tuple of array of float
            A tuple containing two arrays, which hold the x, y, and z
            coordinates of the foci.
        """
        peri = self.xyzPeriastron()
        apas = self.xyzApastron()
        center = (peri + apas) / 2.0
        direc = (peri - apas) / numpy.sqrt(((peri - apas)**2).sum())
        ae = self.a * self.e
        return (center + ae * direc, center - ae * direc)

    def xyzNodes_LOSZ(self, los="+z", getTimes=False):
        """
        Calculate the nodes of the orbit for LOS in +/-z direction.

        The nodes of the orbit are the points at which
        the orbit cuts the plane of the sky. In this case,
        these are the points at which the z-coordinate
        vanishes, i.e., the x-y plane is regarded the plane
        of the sky.

        Parameters
        ----------
        los : string, {+z,-z}, optional
            Line of sight points either in +z direction (observer
            at -z) or vice versa. Changing the direction
            interchanges the ascending and descending node.
        getTimes : boolean, optional
            If True, also the times will be returned at which the
            nodes are reached. Default is False.

        Returns
        -------
        Nodes : Tuple of two coordinate arrays
            Returns the xyz coordinates of both nodes. The first is the
            ascending node and the second is the descending node.
            If `getTimes` is True, tuples of the form (xyz-pos, time)
            will be returned.
        """
        # f = -w (z-component vanishes there)
        E = arctan(tan(-self._w / 2.0) *
                   sqrt((1. - self.e) / (1. + self.e))) * 2.
        M = E - self.e * sin(E)
        t1 = M / self._n + self.tau
        node1 = self.xyzPos(t1)
        # Velocity is used to distinguish nodes
        v1 = self.xyzVel(t1)
        # f = -w + pi
        E = arctan(tan((-self._w + pi) / 2.0) *
                   sqrt((1. - self.e) / (1. + self.e))) * 2.
        M = E - self.e * sin(E)
        t2 = M / self._n + self.tau
        node2 = self.xyzPos(t2)
        # Find the ascending and descending node
        from PyAstronomy.pyasl import LineOfSight
        l = LineOfSight(los).los
        if abs(l[2]) != 1.0:
            raise(PE.PyAValError("Invalid line of sight (LOS): " + str(los),
                                 where="xyzNodes_LOSZ",
                                 solution="Use '-z' or '+z'."))

        if getTimes:
            node1, node2 = (node1, t1), (node2, t2)
        
        if l[2] == 1.0:
            # Looking in +z direction
            if v1[2] > 0.0:
                # First node is ascending
                return (node1, node2)
            else:
                return (node2, node1)
        else:
            # Looking in -z direction
            if v1[2] < 0.0:
                # First node is ascending
                return (node1, node2)
            else:
                return (node2, node1)

    def xyzNodes(self):
        """
        Calculate the nodes of the orbit.

        The nodes of the orbit are the points at which
        the orbit cuts the observing plane. In this case,
        these are the points at which the z-coordinate
        vanishes, i.e., the x-y plane is regarded the plane
        of observation.

        Returns
        -------
        Nodes : Tuple of two coordinate arrays
            Returns the xyz coordinates of both nodes. 
        """
        raise(PE.PyADeprecationError("xyzNodes is deprecated.",
                                     solution="Please use 'xyzNodes_LOSZ' instead."))

    def orbAngMomentum(self, t=0.0):
        """
        The specific orbital angular momentum of a body in the current orbit.

        Parameters
        ----------
        t : float, optional
            The time used to calculate the angular momentum. As
            is it constant for a Kepler orbit, this parameter is
            of no relevance.

        Results
        -------
        Specific angular momentum : array with three elements
            The specific angular momentum (i.e., per unit mass)
            of a body on the current orbit.
        """
        r = self.xyzPos(t)
        v = self.xyzVel(t)
        return numpy.cross(r, v)

    def projPlaStDist(self, t):
        """
        Calculate the sky-projected planet-star separation.

        Parameters
        ----------
        t : float or array
            The time axis.

        Returns
        -------
        Position : array
            The sky-projected planet-star separation at the given time.
            If the input was an array, the output will be an array, 
            holding the separations at the given times.
        """
        p = self.a * (1. - self.e**2)
        E = self._getEccentricAnomaly(t)
        f = arctan(sqrt((1. + self.e) / (1. - self.e)) * tan(E / 2.)) * 2.0
        wf = self._w + f
        if not hasattr(wf, "__iter__"):
            psdist = p / (1. + self.e * cos(f)) * \
                sqrt(1. - sin(self._i)**2 * sin(wf)**2)

        else:
            # Assume it is an array
            psdist = numpy.zeros(len(t))
            for i in smo.range(len(t)):
                psdist[i] = p / (1. + self.e * cos(f[i])) * \
                    sqrt(1. - sin(self._i)**2 * sin(wf[i])**2)

        return psdist

    def yzCrossingTime(self, ordering="y"):
        """
        Calculate times of crossing the yz-plane.

        This method calculates the times at which
        the yz-plane is crossed by the orbit. This
        is equivalent to finding the times where
        x=0.
        
        Parameters
        ----------
        ordering : string, {y,z}
            Determines whether the sign of y or z axis position at
            crossing time is used for ordering the output.

        Returns
        -------
        Time 1 : float
            First crossing time defined as having POSITIVE
            y/z position.
        Time 2 : float
            Second crossing time defined as having NEGATIVE
            y/z position.
        """
        if abs(self._Omega) < 1e-16:
            f = -self._w + pi / 2.0
        else:
            f = -self._w + arctan(1.0 / (tan(self._Omega) * cos(self._i)))
        E = 2. * arctan(sqrt((1 - self.e) / (1. + self.e)) * tan(f / 2.))
        t1 = (E - self.e * sin(E)) / self._n + self.tau
        p1 = self.xyzPos(t1)
        f += pi
        E = 2. * arctan(sqrt((1 - self.e) / (1. + self.e)) * tan(f / 2.))
        t2 = (E - self.e * sin(E)) / self._n + self.tau

        t1 -= self._per * numpy.floor(t1 / self._per)
        t2 -= self._per * numpy.floor(t2 / self._per)

        if not ordering in ["y", "z"]:
            raise(PE.PyAValError("Use 'y' or 'z' for `ordering`."))
        i = {"y":1, "z":2}[ordering]

        if p1[i] >= 0.0:
            # y position of p1 is > 0
            return (t1, t2)
        else:
            return (t2, t1)

    def xzCrossingTime(self):
        """
        Calculate times of crossing the xz-plane.

        This method calculates the times at which
        the xz-plane is crossed by the orbit. This
        is equivalent to finding the times where
        y=0.

        Returns
        -------
        Time 1 : float
            First crossing time defined as having POSITIVE
            x position.
        Time 2 : float
            Second crossing time defined as having NEGATIVE
            x position.
        """
        f = -self._w - arctan(tan(self._Omega) / cos(self._i))
        E = 2. * arctan(sqrt((1 - self.e) / (1. + self.e)) * tan(f / 2.))
        t1 = (E - self.e * sin(E)) / self._n + self.tau
        p1 = self.xyzPos(t1)
        f += pi
        E = 2. * arctan(sqrt((1 - self.e) / (1. + self.e)) * tan(f / 2.))
        t2 = (E - self.e * sin(E)) / self._n + self.tau

        t1 -= self._per * numpy.floor(t1 / self._per)
        t2 -= self._per * numpy.floor(t2 / self._per)

        if p1[0] >= 0.0:
            # y position of p1 is > 0
            return (t1, t2)
        else:
            return (t2, t1)

    def _setPer(self, per):
        self._per = per
        self._n = 2.0 * pi / self._per

    def _seti(self, i):
        self._i = i / 180. * pi

    def _setw(self, w):
        self._w = w / 180. * pi

    def _setOmega(self, omega):
        self._Omega = omega / 180. * pi

    per = property(lambda self: self._per, _setPer, doc="The orbital period.")
    i = property(lambda self: self._i / pi * 180, _seti)
    w = property(lambda self: self._w / pi * 180, _setw)
    Omega = property(lambda self: self._Omega / pi * 180, _setOmega)

    def __init__(self, a, per, e=0, tau=0, Omega=0, w=0, i=0, ks=MarkleyKESolver):
        # i, w, Omega are properties so that the numbers can be given in
        # deg always. The underscored attributes are in rad.
        self.i = i
        self.w = w
        self.Omega = Omega
        self.e = e
        self.a = a
        self.per = per
        self.tau = tau
        self.ks = ks()


class BinaryOrbit(object):
    """
    Keplerian orbits of binary components
    
    The binary system consists of the primary mass (m1) and the secondary mass (m2),
    which orbit a common center of mass. The latter is in the origin of the coordinate system,
    stationary, and identical with a focus of either of the orbit ellipses.
    The system is parameterized by its total mass (m1+m2), the mass ratio (m2/m1), and the orbital
    elements of the primary orbit. The orbit of the secondary has a different semi-major axis
    than that of the primary and its argument of periapsis is shifted by 180 degrees.
    The orientation of the orbit ellipses follows the convention of :class:`KeplerEllipse`
    
    Parameters
    ----------
    m2m1 : float
        Mass ratio between secondary and primary component (e.g., low for planetary systems)
    mtot : float
         Total mass of the system [solar masses]
    per : float
        Orbital period [d]
    e : float, optional
        Orbital eccentricity (0-1).
    tau : float, optional
        Time of periapsis passage [d].
    Omega : float, optional
        Longitude of the ascending node [deg].
    w : float, optional
        Argument of periapsis (primary orbit) [deg]. Note that the longitude
        if periapsis is given by Omega+w.
    i : float, optional
        Orbit inclination [deg].
    ks : Class object, optional
        The solver for Kepler's Equation. Default is the
        `MarkleyKESolver`. Each solver must have a `getE`
        method, which takes either a float or array of float
        of mean anomalies and the eccentricity, and returns
        the associated eccentric anomalies.
    msun : float, optional
        Solar mass [kg] to be adopted. By default the value from PyA's constants
        will be used.    
    """
    
    def _calcSMAs(self):
        """
        Calculate semi-major-axes from total mass and mass ratio
        """
        # a1 + a2 from Kepler III
        self._a_m = cbrt( (self._per_s**2 * self._pc.G * self._mtot_kg ) / (4*pi**2) )
        # Get SMA of secondary body
        self._a2_m = self._a_m / (1+self._m2m1)
        self._a1_m = self._a_m - self._a2_m
        
    a = property(fget=lambda self:self._a_m, fset=None, doc="Total semi-major axis (a1+a2) [m]")
    a1 = property(fget=lambda self:self._a1_m, fset=None, doc="Semi-major axis of primary orbit (a1) [m]")
    a2 = property(fget=lambda self:self._a2_m, fset=None, doc="Semi-major axis of secondary orbit (a2) [m]")
    msun = property(fget=lambda self:self._MSun, fset=None, doc="Adopted solar mass [kg]")
        
    def _setm2m1(self, m2m1):
        """
        Set mass ratio
        """
        if m2m1 < 0:
            raise(PE.PyAValError(f"Mass ratio m2m1 must be >= 0. Current value is {m2m1}.", \
                                 where="BinaryOrbit"))
        self._m2m1 = m2m1
        self._calcSMAs()
        
    def _setmtot(self, mtot):
        """
        Set total mass
        
        Parameters
        ----------
        mtot : float > 0
            Total mass of system [MSun]
        """
        if mtot <= 0:
            raise(PE.PyAValError(f"Total mass must be > 0. Current value is {mtot} MSun.", \
                                 where="BinaryOrbit"))  
        self._mtot_msun = mtot
        self._mtot_kg = mtot * self._MSun
        self._calcSMAs()
        
    def _setper(self, per):
        """
        Set orbital period
        
        Parameters
        ----------
        per : float > 0
            Orbital period [days]
        """
        if per <= 0:
            raise(PE.PyAValError(f"Orbital period must be > 0. Current value is {per} days.", \
                                 where="BinaryOrbit"))              
        self._per_d = per
        self._per_s = self._per_d * 86400
    
    def _settau(self, tau):
        """
        Set orbital period
        
        Parameters
        ----------
        tau : float
            Time of periastron [days]
        """           
        self._tau_d = tau
        self._tau_s = self._tau_d * 86400    
        
    m2m1 = property(fget=lambda self:self._m2m1, fset=_setm2m1, doc="Mass ratio m2/m1 (>=0)")
    mtot = property(fget=lambda self:self._mtot_msun, fset=_setmtot, doc="Total mass of system [MSun]")
    per = property(fget=lambda self:self._per_d, fset=_setper, doc="Orbital period [days]")
    tau = property(fget=lambda self:self._tau_d, fset=_settau, doc="Time of periapsis [days]")
            
    def __init__(self, m2m1, mtot, per, e=0, tau=0, Omega=0, w=0, i=0, msun=None, ks=MarkleyKESolver):
        self._pc = _PC.PyAConstants()
        self._pc.setSystem("SI")
        if msun is None:
            self._MSun = self._pc.MSun
        else:
            self._MSun = msun
        
        # Dummies to be replaced shortly
        self._m2m1 = numpy.nan
        self._mtot_kg = numpy.nan
        
        self.per = per
        self.mtot = mtot
        self.m2m1 = m2m1
        self.tau = tau
        
        self._ke = KeplerEllipse(self._a1_m, self._per_s, e=e, tau=self._tau_s, Omega=Omega, w=w, i=i, ks=ks)
        
        # Forward properties to class API
        self.i = self._ke.i
        self.w = self._ke.w
        self.Omega = self._ke.Omega
        
    def getKeplerEllipse_primary(self):
        """
        Get Kepler ellipse model for the primary component
        
        Returns
        -------
        ke : KeplerEllipse
            KeplerEllipse object pertaining to primary object.
        """
        return self._ke

    def getKeplerEllipse_secondary(self):
        """
        Get Kepler ellipse model for the secondary component
        
        Returns
        -------
        ke : KeplerEllipse
            KeplerEllipse object pertaining to secondary object.
        """
        k = copy.deepcopy(self._ke)
        k.w = k.w + 180.
        k.a = self._a2_m
        return k
    
    def xyzPos(self, t):
        """
        Calculate orbit position.

        Parameters
        ----------
        t : float or array
            The time axis [s]

        Returns
        -------
        Position primary : array
            The x, y, and z coordinates of the body at the given time [m].
            If the input was an array, the output will be an array of
            shape (input-length, 3), holding the positions at the given
            times.
        Position secondary : array
            The x, y, and z coordinates of the secondary [m].
        """
        r1 = self._ke.xyzPos(t)
        r2 = -r1/self.m2m1
        return r1, r2
    
    def xyzVel(self, t):
        """
        Calculate orbit velocity.

        Parameters
        ----------
        t : float or array
            The time axis [s].

        Returns
        -------
        Velocity primary : array
            The x, y, and z components of the body's velocity at the
            given time [m/s]. If the input was an array, the output will be
            an array of shape (input-length, 3), holding the velocity
            components at the given times. The unit is that of the
            semi-major axis divided by that of the period.
        Velocity secondary : array
            The x, y, and z velocity components of the secondary [m/s].
        """
        v1 = self._ke.xyzVel(t)
        v2 = -v1/self.m2m1
        return v1, v2
    

def phaseAngle(pos, los='-z'):
    """
    Calculate the phase angle.

    The phase angle is the angle between the star and the Earth (or Sun)
    as seen from the planet (e.g., Seager et al. 1999, ApJ, 504).
    The range of the phase angle is 0 - 180 degrees. In the calculations,
    it is assumed to be located at the center of the coordinate system.

    Parameters
    ----------
    pos : array
        Either a one-dimensional array with xyz coordinate or a
        [3,N] array containing N xyz positions.
    los : LineOfSight object
        A `LineOfSight` object from the pyasl giving the line of
        sight.

    Returns
    -------
    Phase angle : The phase angle in degrees. Depending on the input,
        it returns a single value or an array. 
    """
    from PyAstronomy.pyasl import LineOfSight
    l = LineOfSight(los)
    if pos.shape == (3,):
        # It is a single value
        return numpy.arccos(numpy.sum(-pos * (-l.los)) /
                            numpy.sqrt(numpy.sum(pos**2))) / numpy.pi * 180.
    else:
        # It is an array of positions
        N = len(pos[::, 0])
        result = numpy.zeros(N)
        for i in smo.range(N):
            print(i, numpy.sum((-pos[i, ::]) * (-l.los)), pos[i, ::])
            result[i] = numpy.arccos(numpy.sum((-pos[i, ::]) * (-l.los)) /
                                     numpy.sqrt(numpy.sum(pos[i, ::]**2)))
        return result / numpy.pi * 180.



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


class _Gdl:

    def __init__(self, vsini, epsilon):
        """
        Calculate the broadening profile.

        Parameters
        ----------
        vsini : float
            Projected rotation speed of the star [km/s]
        epsilon : float
            Linear limb-darkening coefficient
        """
        self.vc = vsini / 299792.458
        self.eps = epsilon

    def gdl(self, dl, refwvl, dwl):
        """
        Calculates the broadening profile.

        Parameters
        ----------
        dl : array
            'Delta wavelength': The distance to the reference point in
            wavelength space [A].
        refwvl : array
            The reference wavelength [A].
        dwl : float
            The wavelength bin size [A].

        Returns
        -------
        Broadening profile : array
            The broadening profile according to Gray. 
        """
        self.dlmax = self.vc * refwvl
        self.c1 = 2.*(1. - self.eps) / \
            (np.pi * self.dlmax * (1. - self.eps/3.))
        self.c2 = self.eps / (2. * self.dlmax * (1. - self.eps/3.))
        result = np.zeros(len(dl))
        x = dl/self.dlmax
        indi = np.where(np.abs(x) < 1.0)[0]
        result[indi] = self.c1 * \
            np.sqrt(1. - x[indi]**2) + self.c2*(1. - x[indi]**2)
        # Correct the normalization for numeric accuracy
        # The integral of the function is normalized, however, especially in the case
        # of mild broadening (compared to the wavelength resolution), the discrete
        # broadening profile may no longer be normalized, which leads to a shift of
        # the output spectrum, if not accounted for.
        result /= (np.sum(result) * dwl)
        return result


def rotBroad(wvl, flux, epsilon, vsini, edgeHandling="firstlast"):
    """
    Apply rotational broadening to a spectrum.

    This function applies rotational broadening to a given
    spectrum using the formulae given in Gray's "The Observation
    and Analysis of Stellar Photospheres". It allows for
    limb darkening parameterized by the linear limb-darkening law.

    The `edgeHandling` parameter determines how the effects at
    the edges of the input spectrum are handled. If the default
    option, "firstlast", is used, the input spectrum is internally
    extended on both sides; on the blue edge of the spectrum, the
    first flux value is used and on the red edge, the last value
    is used to extend the flux array. The extension is neglected
    in the return array. If "None" is specified, no special care
    will be taken to handle edge effects.

    .. note:: Currently, the wavelength array as to be regularly
              spaced.

    Parameters
    ----------
    wvl : array
        The wavelength array [A]. Note that a
        regularly spaced array is required.
    flux : array
        The flux array.
    vsini : float
        Projected rotational velocity [km/s].
    epsilon : float
        Linear limb-darkening coefficient (0-1).
    edgeHandling : string, {"firstlast", "None"}
        The method used to handle edge effects.

    Returns
    -------
    Broadened spectrum : array
        An array of the same size as the input flux array,
        which contains the broadened spectrum.
    """
    # Check whether wavelength array is evenly spaced
    sp = wvl[1::] - wvl[0:-1]
    if abs(max(sp) - min(sp)) > 1e-6:
        raise(PE.PyAValError("Input wavelength array is not evenly spaced.",
                             where="pyasl.rotBroad",
                             solution=["Use evenly spaced input array.",
                                       "Use pyasl.equidistantInterpolation to produce equidistantly sampled data."]))
    if vsini <= 0.0:
        raise(PE.PyAValError("vsini must be positive.", where="pyasl.rotBroad"))
    if (epsilon < 0) or (epsilon > 1.0):
        raise(PE.PyAValError("Linear limb-darkening coefficient, epsilon, should be '0 < epsilon < 1'.",
                             where="pyasl.rotBroad",
                             solution="Adapt epsilon."))

    # Wavelength binsize
    dwl = wvl[1] - wvl[0]

    # Indices of the flux array to be returned
    validIndices = None

    if edgeHandling == "firstlast":
        # Number of bins additionally needed at the edges
        binnu = int(np.floor(((vsini / 299792.458) * max(wvl)) / dwl)) + 1
        # Defined 'valid' indices to be returned
        validIndices = np.arange(len(flux)) + binnu
        # Adapt flux array
        front = np.ones(binnu) * flux[0]
        end = np.ones(binnu) * flux[-1]
        flux = np.concatenate((front, flux, end))
        # Adapt wavelength array
        front = (wvl[0] - (np.arange(binnu) + 1) * dwl)[::-1]
        end = wvl[-1] + (np.arange(binnu) + 1) * dwl
        wvl = np.concatenate((front, wvl, end))
    elif edgeHandling == "None":
        validIndices = np.arange(len(flux))
    else:
        raise(PE.PyAValError("Edge handling method '" + str(edgeHandling) + "' currently not supported.",
                             where="pyasl.rotBroad",
                             solution="Choose ones of the valid edge handling methods"))

    result = np.zeros(len(flux))
    gdl = _Gdl(vsini, epsilon)

    for i in smo.range(len(flux)):
        dl = wvl[i] - wvl
        g = gdl.gdl(dl, wvl[i], dwl)
        result[i] = np.sum(flux * g)
    result *= dwl

    return result[validIndices]


def fastRotBroad(wvl, flux, epsilon, vsini, effWvl=None):
    """
    Apply rotational broadening using a single broadening kernel.

    The effect of rotational broadening on the spectrum is
    wavelength dependent, because the Doppler shift depends
    on wavelength. This function neglects this dependence, which
    is weak if the wavelength range is not too large.

    .. note:: numpy.convolve is used to carry out the convolution
              and "mode = same" is used. Therefore, the output
              will be of the same size as the input, but it
              will show edge effects.

    Parameters
    ----------
    wvl : array
        The wavelength
    flux : array
        The flux
    epsilon : float
        Linear limb-darkening coefficient
    vsini : float
        Projected rotational velocity in km/s.
    effWvl : float, optional
        The wavelength at which the broadening
        kernel is evaluated. If not specified,
        the mean wavelength of the input will be
        used.

    Returns
    -------
    Broadened spectrum : array
        The rotationally broadened output spectrum.
    """
    # Check whether wavelength array is evenly spaced
    sp = wvl[1::] - wvl[0:-1]
    if abs(max(sp) - min(sp)) > 1e-6:
        raise(PE.PyAValError("Input wavelength array is not evenly spaced.",
                             where="pyasl.fastRotBroad",
                             solution=["Use evenly spaced input array.",
                                       "Use pyasl.equidistantInterpolation to produce equidistantly sampled data."]))
    if vsini <= 0.0:
        raise(PE.PyAValError("vsini must be positive.", where="pyasl.rotBroad"))
    if (epsilon < 0) or (epsilon > 1.0):
        raise(PE.PyAValError("Linear limb-darkening coefficient, epsilon, should be '0 < epsilon < 1'.",
                             where="pyasl.rotBroad",
                             solution="Adapt epsilon."))

    # Wavelength binsize
    dwl = wvl[1] - wvl[0]

    if effWvl is None:
        effWvl = np.mean(wvl)

    gdl = _Gdl(vsini, epsilon)

    # The number of bins needed to create the broadening kernel
    binnHalf = int(np.floor(((vsini / 299792.458) * effWvl / dwl))) + 1
    gwvl = (np.arange(4*binnHalf) - 2*binnHalf) * dwl + effWvl
    # Create the broadening kernel
    dl = gwvl - effWvl
    g = gdl.gdl(dl, effWvl, dwl)
    # Remove the zero entries
    indi = np.where(g > 0.0)[0]
    g = g[indi]

    result = np.convolve(flux, g, mode="same") * dwl
    return result



from PyAstronomy.pyaC import pyaErrors as PE
import numpy as np


def magToFluxDensity_bessel98(band, mag, mode="nu"):
    """
    Convert magnitude into flux density according to Bessel et al. 1998

    The conversion implemented here is based on the data given in Table A2 of
    Bessel et al. 1998, A&A 333, 231-250, which gives "Effective wavelengths (for an A0 star),
    absolute fluxes (corresponding to zero magnitude) and zeropoint magnitudes for the UBVRI-
    JHKL Cousins-Glass-Johnson system". Note that zp(f_nu) and zp(f_lam) are exchanged in
    the original table.

    Parameters
    ----------
    band : string
        Any of U, B, V, R, I, J, H, K, Kp, L, and L*
    mag : float, array
        The magnitude value to be converted
    mode : string, {nu, mod}
        Determines whether f_nu or f_lam will be calculated.

    Returns
    -------
    f_nu/lam : float
        The corresponding flux density in units if erg/cm**2/s/Hz in the
        case of mode 'nu' and erg/cm**2/s/A in the case of 'lam'.
    lam_eff : float
        Effective filter wavelength in Angstrom
    """
    if not mode in ["nu", "lam"]:
        raise (
            PE.PyAValError(
                "Unknown mode",
                where="magToFluxDensity_bessel98",
                solution="Use 'nu' or 'lam'",
            )
        )

    # Data from Bessel 1998 Table A2
    d = """bands U B V R I J H K Kp L L*
    lam_eff 0.366 0.438 0.545 0.641 0.798 1.22 1.63 2.19 2.12 3.45 3.80
    f_nu 1.790 4.063 3.636 3.064 2.416 1.589 1.021 0.640 0.676 0.285 0.238
    f_lam 417.5 632 363.1 217.7 112.6 31.47 11.38 3.961 4.479 0.708 0.489
    zp(f_nu) 0.770 -0.120 0.000 0.186 0.444 0.899 1.379 1.886 1.826 2.765 2.961
    zp(f_lam) -0.152 -0.602 0.000 0.555 1.271 2.655 3.760 4.906 4.780 6.775 7.177"""
    # Digest table
    s = d.split()
    t = {}
    for i in range(6):
        t[s[i * 12]] = s[i * 12 + 1 : (i + 1) * 12]
        if s[i * 12] != "bands":
            t[s[i * 12]] = np.array(t[s[i * 12]], dtype=float)

    if not band in t["bands"]:
        raise (
            PE.PyAValError(
                "Unknown passband: " + str(band),
                where="magToFluxDensity_bessel98",
                solution="Use any of: " + ", ".join(t["bands"]),
            )
        )

    # Band index
    bi = t["bands"].index(band)

    c = {"nu": 48.598, "lam": 21.1}

    f = 10.0 ** ((mag + c[mode] + t["zp(f_" + mode + ")"][bi]) / -2.5)
    return f, t["lam_eff"][bi] * 1e4


def absMagToPower(am, absMagSun=4.75, absLumSun=3.846e33):
    """
    Convert absolute magnitude to power scale

    The default values for the absolute magnitude and luminosity
    of the Sun are adopted from Harmanec and Prsa 2011
    (2011PASP..123..976H).

    Parameters
    ----------
    am : float
        Absolute magnitude.
    absMagSun : float, optional
        Absolute magnitude of the Sun.
    absLumSun : float, optional
        Absolute luminosity of the Sun.
        The default is given in units of erg/s.

    Returns
    -------
    Power : float
        Total emitted power. Same units as `absLumSun`;
        the default corresponds to erg/s.
    """
    power = 10.0 ** ((am - absMagSun) / (-2.5)) * absLumSun
    return power


def absModuleToDist(magApp, magAbs):
    """
    Convert apparent and absolute magnitude into distance.

    Parameters
    ----------
    magApp : float
        Apparent magnitude of object.
    magAbs : float
        Absolute magnitude of object.

    Returns
    -------
    Distance : float
        The distance resulting from the difference in
        apparent and absolute magnitude [pc].
    """
    d = 10.0 ** (-(magAbs - magApp) / 5.0 + 1.0)
    return d



# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy
from .onedfit import OneDFit


class CauchyLorentz1d(OneDFit):
    r"""
    A Cauchy-Lorentz profile

    *Fit parameters*:
     - `A` - Amplitude
     - `g` - Scale parameter (usually gamma)
     - `mu` - Center
     - `off` - A constant offset
     - `lin` - A linear contribution
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "g", "mu", "off", "lin"])
        self.setRootName("CauLor")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        y = self["A"] * (self["g"]/numpy.pi / ((x-self["mu"])**2 + self["g"]**2)) + \
            (self["lin"] * x) + self["off"]
        return y



from __future__ import print_function, division
import copy
import re
import numpy
from PyAstronomy.pyaC import pyaErrors as PE
from .onedfit import MiniFunc
from .params import equal
from .onedfit import _PyMCSampler, _OndeDFitParBase
from .nameIdentBase import ModelNameIdentBase
from PyAstronomy import pyaC
import six
import six.moves as smo

from PyAstronomy.funcFit import _scoImport, _pymcImport
if _scoImport:
    import scipy.optimize as sco
if _pymcImport:
    import pymc


class MiniFuncSync:
    r"""
    This decorator can be applied to use
    self-defined objective functions.

    Applied to an objective function, it adds the functionality
    needed to evaluate the model given a certain parameter vector,
    so that the user does only have to take care
    about the quantity to be minimized.

    Parameters
    ----------
    odf : fitting object
        The fitting object that is supposed to use the self-defined
        objective function.
    """

    def __init__(self, odf):
        r"""
        Parameter:
          - `odf` - An instance of a fitting objects such as for example *GaussFit1d*.
        """
        # Save a REFERENCE to the fitting object
        self.odf = odf

    def __call__(self, f):
        r"""
        Parameter:
         - `f` - The user-defined objective function.
        """
        def miniFunc(P):
            # Update the parameter values in the 'Params' class instance.
            self.odf.pars.setFreeParams(P)
            # Update self.model to hold the evaluated function.
            self.odf.updateModel()

            val = f(self.odf, P)
            # Assign penalty
            val += self.odf.pars.getPenalty(
                penaltyFact=self.odf.penaltyFactor)[0]
            val += self.odf.pars.applyConditionalRestrictions()

            return val
        return miniFunc


class SyncFitContainer(_PyMCSampler, _OndeDFitParBase):

    def addComponent(self, newCompo):
        r"""
        Add a new component to the container.

        Parameters:
          - `newCompo` - A funcFit model.

        Returns:
          The component identifier.
        """
        # Copy component into internal model container
        # (use 1 for the first such model)
        n = len(self._compos) + 1
        self._compos[n] = copy.deepcopy(newCompo)
        # Rename the parameters (attach '_sn' with n being the component number)
        for c in self._compos[n]._compoWalk():
            if c._isComposed():
                continue
            c.setRootName(c.naming.getRoot()+"[s"+str(n)+"]", rename=True)
        # Update the pars member
        if self.pars is not None:
            self.pars = self.pars + self._compos[n].pars
        else:
            self.pars = copy.deepcopy(self._compos[n].pars)
        # Assign reference to updated parameter set (in ALL models)
        for k in six.iterkeys(self._compos):
            self._compos[k].pars = self.pars
            self._compos[k]._newPars(self.pars)
        # Define the propMap (identical mapping)
        self.propMap = dict(
            zip(self.parameters().keys(), self.parameters().keys()))
        # Returns the component identifier
        return n

    def components(self):
        r"""
        Returns:
          A list holding the component names.
        """
        return list(self._compos.keys())

    def evaluate(self, axes, component=None):
        r"""
        Parameters:
          - `axes` - If `component` is not given, a dictionary holding the x-axis for each \
                     component name. Otherwise, the x-axis for the specified component.
          - `component` - string, optional, The name of the component to be evaluated.

        The evaluated model(s) is saved in the `models` dictionary.
        """
        if component is not None:
            self.models[component] = self._compos[component].evaluate(axes)
        else:
            for k, v in six.iteritems(axes):
                self.models[k] = self._compos[k].evaluate(v)

    def updateModel(self):
        r"""
        Evaluate all components. Updates the values in the `models` dictionary.
        """
        for c in six.iterkeys(self._compos):
            self.evaluate(self.data[c][0], component=c)

    def __chiSqr(self):
        @MiniFuncSync(self)
        def miniChiSqr(odf, P):
            # Calculate chi^2 and apply penalty if boundaries are violated.
            chi = 0.0
            for k in six.iterkeys(self._compos):
                chi += numpy.sum(((self.data[k][1] -
                                 self.models[k])/self.yerr[k])**2)
            return chi
        return miniChiSqr

    def __sqrDiff(self):
        @MiniFuncSync(self)
        def minisqr(odf, P):
            # Calculate squared difference
            sqr = 0.0
            for k in six.iterkeys(self._compos):
                sqr += numpy.sum((self.data[k][1] - self.models[k])**2)
            return sqr
        return minisqr

    def __cash79(self):
        @MiniFuncSync(self)
        def miniCash79(odf, P):
            # Calculate Cash statistics according to Cash 1979 (ApJ 228, 939)
            cc = 0
            for k in six.iterkeys(self._compos):
                cc += -2.0 * \
                    numpy.sum(
                        self.data[k][1] * numpy.log(self.models[k]) - self.models[k])
            return cc
        return miniCash79

    def __chiSqrRobust(self):
        @MiniFuncSync(self)
        def miniChiSqr(odf, P):
            # Calculate chi^2 and apply penalty if boundaries are violated.
            chi = 0.0
            for k in six.iterkeys(self._compos):
                chi += numpy.nansum(((self.data[k][1] -
                                    self.models[k])/self.yerr[k])**2)
            return chi
        return miniChiSqr

    def __sqrDiffRobust(self):
        @MiniFuncSync(self)
        def minisqr(odf, P):
            # Calculate squared difference
            sqr = 0.0
            for k in six.iterkeys(self._compos):
                sqr += numpy.nansum((self.data[k][1] - self.models[k])**2)
            return sqr
        return minisqr

    def __cash79Robust(self):
        @MiniFuncSync(self)
        def miniCash79(odf, P):
            # Calculate Cash statistics according to Cash 1979 (ApJ 228, 939)
            cc = 0
            for k in six.iterkeys(self._compos):
                cc += -2.0 * \
                    numpy.nansum(
                        self.data[k][1] * numpy.log(self.models[k]) - self.models[k])
            return cc
        return miniCash79

    def treatAsEqual(self, parameter):
        r"""
        Treat parameters as equal.

        `parameter` - string or list of string,
                      If a string is given, all parameters with this "base name" (i.e., neglecting
                      everything after an underscore) will be treated as equal.
                      Otherwise the specified parameters will be treated as equal.

        This method uses the *relations* known from *OneDFit* to treat parameters as equal.
        Dependent variables are thawed before the relation is applied, if they are not already
        free.
        """
        if isinstance(parameter, six.string_types):
            ps = []
            for p in six.iterkeys(self.parameters()):
                r = re.match(parameter+"_.*", p)
                if r is not None:
                    ps.append(p)
            ps = sorted(ps)
        else:
            ps = parameter
        for i in smo.range(1, len(ps)):
            if not ps[i] in self.freeParamNames():
                self.thaw(ps[i])
            self.relate(ps[i], [ps[0]], equal)

    def parameterSummary(self, toScreen=True, prefix=""):
        lines = []
        for k, v in six.iteritems(self._compos):
            lines.append(prefix)
            lines.append(
                prefix + "Parameters for syncFit Component: " + str(k))
            lines.append(prefix + "=" * len(lines[-1]))
            lines.extend(v.parameterSummary(toScreen=False, prefix=prefix))
        if toScreen:
            for l in lines:
                print(l)
        return lines

    def setObjectiveFunction(self, miniFunc="chisqr"):
        r"""
        Define the objective function.

        This function sets the `miniFunc` attribute, which is used
        to calculate the quantity to be minimized.

        Parameters
        ----------
        miniFunc : str {chisqr, cash79, sqrdiff} or callable
            The objective function. If "chisqr", chi-square will be
            minimzed. If "cash 79", the Cash statistics 
            (Cash 1979, ApJ 228, 939, Eq. 5) will be used.
            If "sqrdiff" is specified, 
            Otherwise, a user-defined function is assumed.
        """
        # Determine function to be minimized
        if miniFunc == "chisqr":
            self.miniFunc = self.__chiSqr()
            return
        elif miniFunc == "cash79":
            self.miniFunc = self.__cash79()
            return
        elif miniFunc == "sqrdiff":
            self.miniFunc = self.__sqrDiff()
            return
        elif miniFunc == "chisqrRobust":
            self.miniFunc = self.__chiSqrRobust()
            return
        elif miniFunc == "cash79Robust":
            self.miniFunc = self.__cash79Robust()
            return
        elif miniFunc == "sqrdiffRobust":
            self.miniFunc = self.__sqrDiffRobust()
            return
        else:
            if not hasattr(miniFunc, '__call__'):
                raise (PE.PyAValError("`miniFunc` is neither None, a valid string, or a function.",
                                      where="OneDFit::fit",
                                      solution="Use, e.g., 'chisqr' or another valid choice from the documentation."))

            # A function has been specified
            self.miniFunc = miniFunc
            return

    def fit(self, data, yerr=None, X0=None, minAlgo=None, miniFunc=None, *fminPars, **fminArgs):
        r"""
        Carries out the fit.

        In principle, any fit algorithm can be used. If none is specified, the default is \
        scipy.optimize.fmin (Nelder-Mead Simplex). Another choice could for instance be \
        scipy.optimize.fmin_powell. After the fit, the return value of the fitting method \
        is stored in the class property `fitResult` and the `model` property is set to \
        the best fit.

        Parameters:
          - `data` - Dictionary of the form: {c:[x, y], ...}. Here `c` is the component number (starts
                     with one, and `x`, `y` are the x-axis and associated values. 
          - `yerr` - array, optional: Error of data values. A dictionary of the from: {c:yerr}, where
                     `c` is the component and yerr the array of error bars.
          - `X0`   - list, optional: The initial guess. If not provided, it will be assumed \
                   that self.pars already contains the initial guess.
          - `minAlgo` - callable, The minimization algorithm. Default is scipy.optimize.fmin; \
                        other algorithms from scipy may be chosen. Alternatively, any callable \
                        object taking the function to minimize as the first, the vector of starting \
                        values as the second, and a full_output flag as arguments can be used.
          - `fminArgs` - Keywords which are passed to the minimization method (default is \
                         of scipy.optimize.fmin) (e.g., `xtol` or `ftol`).
          - `fminPars` - Non-keyword arguments passed to the  minimization method (e.g., fprime in \
                         scipy.optimize.fmin_ncg).
        """
        # Assign internal data properties.

        if data is not None:
            self.data = data
        elif self.data is not None:
            data = self.data
        else:
            raise (PE.PyAValError("You must provide data to fit.",
                   solution="Call fit with data."))

        if yerr is not None:
            self.yerr = yerr
        # Choose minimization algorithm
        if minAlgo is None:
            # If not specified use default.
            # Check whether it is available...
            global _scoImport
            if not _scoImport:
                raise (PE.PyARequiredImport("SciPy.optimize could not be imported.",
                       solution="Install SciPy (see www.scipy.org/)."))
            self.minAlgo = sco.fmin
        else:
            self.minAlgo = minAlgo
        # Determine function to be minimized
        if (miniFunc is None) and (yerr is not None):
            miniFunc = "chisqr"
        elif (miniFunc is None) and (yerr is None):
            miniFunc = "sqrdiff"
        self.setObjectiveFunction(miniFunc)
        # Assign initial guess if necessary
        if X0 is not None:
            self.pars.setFreeParams(X0)
        # Save fminPars and fminArgs to internal variables
        self.fminArgs = fminArgs
        self.fminPars = fminPars
        # Carry out fit
        self.fitResult = self.minAlgo(self.miniFunc, self.pars.getFreeParams(), *self.fminPars,
                                      full_output=True, **self.fminArgs)
        self.pars.setFreeParams(self.fitResult[0])
        self.updateModel()
        self._stepparEnabled = True

    def fitMCMC(self, data, X0, Lims, Steps, yerr=None, pymcPars=None, pyy=None,
                potentials=None, dbfile="mcmcSample.tmp", dbArgs=None, adaptiveMetropolis=False,
                **sampleArgs):
        r"""
        Carry out MCMC fit/error estimation.

        This member is designed to provide a flexible but *easy to use* interface to \
        the capabilities of pymc. In the simplest case, it assumes a Poisson or Gaussian \
        distribution of data points and uses continuous, uniform variables (all free fitting \
        variables) with starting values defined by `X0`, Limits given by `Lims`, and
        step sizes given by `Steps` to sample from the posterior.

        .. note::
           The result (the Marchov-Chain/PyMC MCMC-object) will \
           be contained in the self.MCMC property; the output dictionary of MCMC.stats() \
           (Mean, HPD interval etc.) is saved to self.basicStats.

        Parameters:
          - `x` - An array providing the x-values of the data points.
          - `y` - An array providing the y-values of the data points. \
                  Note that for the MCMC process, it is essential to know the underlying \
                  distribution of the data points. *fitMCMC* assumes Poisson distributed data \
                  of `yerr` is not specified and Gaussian data if it is specified. If other \
                  distributions shall be used, the `pyy` parameter must contain a \
                  *pymc* random variable specifying it.
          - `yerr` - array, optional,
                     Error of data values. A dictionary of the from: {c:yerr}, where
                     `c` is the component and yerr the array of error bars. If specified a Gaussian
                     distribution will be assumed for the data points, otherwise a Poisson distribution
                     is assumed.
          - `pyy` - *optional*,
                    Can be used to handle a PyMC variable containing the \
                    data. This can be useful if the distribution is neither Poisson nor \
                    Gaussian as otherwise assumed by this function.
          - `X0` - A dictionary holding {"parName":value, ...} specifying the start values. \
                   Note that parameters treated in pymcPars must not be part of this.
          - `Lims` - A dictionary of type {"ParName:[min,max], ...} specifying the lower \
                     and upper limit of a variable. \
                     Note that parameters treated in pymcPars must not be part of this.
          - `Steps` - A dictionary providing the step sizes for the MCMC \
                      sampler.
          - `pymcPars` - *optional*,
                         This variable is supposed to hold a dictionary \
                         of the form {"parName":PyMC-Variable, ...}. `pymcPars` can be \
                         used to specify a nonuniform distribution for a parameter.
          - `potentials` - *optional*,
                           Can be used to provide a list of PyMC potentials,
                           which may be needed to provide priors.
          - `dbfile` - The name of the output file, which is to hold the MCMC chain \
                       produced during sampling.
          - `**sampleArgs` - *optional*,
                             Here additional keywords can be specified, which \
                             will be handed to the *isample* member of PyMC. Most notably \
                             it is useful to specify **iter**, **burn**, and **thin**. For other \
                             possibilities see PyMC documentation.
        """
        global _pymcImport
        if not _pymcImport:
            raise (PE.PyARequiredImport("pymc package could not be imported.",
                   solution="Install pymc (see http://code.google.com/p/pymc/"))
        # Assign mutable default parameters
        if pymcPars is None:
            pymcPars = {}
        if dbArgs is None:
            dbArgs = {}
        if potentials is None:
            potentials = []
        self.data = data
        self.yerr = yerr
        # Build up "concatenated" y-axis and yerr axis
        self.ycon = None
        self.yerrcon = None
        for k in six.iterkeys(self._compos):
            if self.ycon is None:
                self.ycon = self.data[k][1].copy()
            else:
                self.ycon = numpy.concatenate((self.ycon, self.data[k][1]))
            if (self.yerr is not None):
                if (self.yerrcon is None):
                    self.yerrcon = self.yerr[k].copy()
                else:
                    self.yerrcon = numpy.concatenate(
                        (self.yerrcon, self.yerr[k]))
        # Copy the pymcPars dictionary (prevents error on multiple sampler calls)
        pymcPars = pymcPars.copy()
        # Get the names of the free parameters
        freeNames = self.freeParamNames()
        print("Free parameters: ", freeNames)
        # Check whether parameter lists are complete, define default steps
        # if necessary.
        self._dictComplete(freeNames, X0, "start values",
                           forget=list(pymcPars))
        self._dictComplete(freeNames, Lims, "limits", forget=list(pymcPars))
        self._dictComplete(freeNames, Steps, "steps")

        # Define (or complete) the pymcPars dictionary by defining uniformly distributed
        # variables in the range [lim[0], lim[1]] with starting values defined by X0.
        for par in freeNames:
            if par in pymcPars:
                continue
            print("Using uniform distribution for parameter: ", par)
            print("  Start value: ",
                  X0[par], ", Limits = [", Lims[par][0], ", ", Lims[par][1], "]")
            pymcPars[par] = pymc.Uniform(
                par, lower=Lims[par][0], upper=Lims[par][1], value=X0[par], doc="Automatically assigned parameter.")

        def getConcatenatedModel():
            result = None
            for k in six.iterkeys(self._compos):
                if result is None:
                    result = self.models[k]
                else:
                    result = numpy.concatenate((result, self.models[k]))
            return result

        # This function is used to update the model
        def getModel(**vals):
            self.assignValue(vals)
            self.updateModel()
            return getConcatenatedModel()

        modelDet = pymc.Deterministic(
            eval=getModel,
            name='Model',
            parents=pymcPars,
            doc='The model',
            trace=True,
            verbose=0,
            dtype=float,
            plot=False,
            cache_depth=2)

        # Define the 'data' (y-values)
        if pyy is None:
            if yerr is None:
                print(
                    "Assuming Poisson distribution for 'y'. Use 'pyy' parameter to change this!")
                pyy = pymc.Poisson(
                    "y", mu=modelDet, value=self.ycon, observed=True)
            else:
                print(
                    "Assuming Gaussian distribution for 'y'. Use 'pyy' parameter to change this!")
                pyy = pymc.Normal("y", mu=modelDet, tau=1.0 /
                                  self.yerrcon**2, value=self.ycon, observed=True)

        # Add data to the Model
        Model = [pyy]
        # Add potentials (e.g., priors)
        Model.extend(potentials)
        # Add free parameters
        for v in six.itervalues(pymcPars):
            Model.append(v)

        # Check database arguments
        if not "dbname" in dbArgs:
            dbArgs["dbname"] = dbfile
        dbArgs = self._checkDbArgs(dbArgs)

        print("Using database arguments: ", dbArgs)
        self.MCMC = pymc.MCMC(Model, **dbArgs)

        # Tell the MCMC class to use the MH algorithm with specified step width
        if adaptiveMetropolis:
            self.MCMC.use_step_method(pymc.AdaptiveMetropolis, list(
                pymcPars.values()), shrink_if_necessary=True)
        else:
            for par in six.iterkeys(pymcPars):
                self.MCMC.use_step_method(
                    pymc.Metropolis, pymcPars[par], proposal_sd=Steps[par], proposal_distribution='Normal')

        if not "iter" in sampleArgs:
            sampleArgs["iter"] = 2000
        if not "burn" in sampleArgs:
            sampleArgs["burn"] = 0
        if not "thin" in sampleArgs:
            sampleArgs["thin"] = 1

        print("Giving the following arguments to 'isample':")
        print("  ", sampleArgs)

        self.MCMC.isample(**sampleArgs)
        self.basicStats = self.MCMC.stats()
        self._basicStatMCMCOutput(self.basicStats)

        # Setting values to ``best fit values'' (lowest deviance)
        mindex = numpy.argmin(self.MCMC.trace("deviance")[:])
        for par in six.iterkeys(pymcPars):
            self[par] = self.MCMC.trace(par)[mindex]
        self.updateModel()
        self.MCMC.db.close()

    def __extractFunctionValue(self, fr):
        r"""
        Returns the function value (e.g., chi-square).

        Parameters
        ----------
        fr : list
            The fit result returned by the fit method
            used by the `fit` method.

        Returns
        -------
        Function value : float
            For example, chi-square.
        """
        return fr[1]

    def steppar(self, pars, ranges, extractFctVal=None, quiet=False):
        r"""
        Allows to step a parameter through a specified range.

        This function steps the specified parameters through the given
        ranges. During each steps, all free parameters, except for those
        which are stepped, are fitted. The resulting contours allow
        to estimate confidence intervals.

        This command uses the fitting parameters specified on a call
        to the `fit` method. In particular, the same values for `x`,
        `y`, `yerr`, `minAlgo`, `miniFunc`, `fminPars`, and `fminArgs`
        are used.

        .. note:: You need to have carried out a fit before you can
                  use `steppar`.

        Parameters
        ----------
        pars : string or list of strings
            The parameter(s) which are to be stepped.
        ranges : dictionary
            A dictionary mapping parameter name to range specifier.
            The latter is a list containing [lower limit, upper limit,
            no. of steps, 'lin'/'log']. The fourth entry, which
            is optional, is a string specifying whether a constant
            linear step size ('lin') or a constant logarithmic
            step size ('log') shall be used.
        quiet : boolean, optional
            If True, output will be suppressed.
        extractFctVal : callable, optional
            A function specifying how the function value is extracted
            from the fit result. If standard settings are used, the
            default of None is adequate.

        Returns
        -------
        Parameter steps : list
            The return value is a list of lists. Each individual list
            contains the values of the stepped parameters as the first
            entries (same order as the input `pars` list), the
            following entry is the value of the objective function
            (e.g., chi square), and the last entry is a tuple
            containing the indices of the steps of the parameter values.
            This last entry can be useful to convert the result into
            an arrow to plot, e.g., contours. 
        """
        if not self._stepparEnabled:
            raise (PE.PyAOrderError("Before you can use steppar, you must call a function, which enables its use (e.g., `fit`).",
                                    solution="Call the `fit` method first and then try again."))
        if isinstance(pars, six.string_types):
            # Make it a list
            pars = [pars]
        # Check parameter consistency
        for p in pars:
            # Check existence
            tmp = self[p]
            if not p in ranges:
                raise (PE.PyAValError("There is no range for parameter: " + p,
                                      solution="Specify a range; e.g., {'xyz':[0.5,1.9,20,'lin']}"))
        # Function to extract function value from the fit result
        if extractFctVal is None:
            self._extractFctVal = self.__extractFunctionValue
        else:
            if not hasattr(extractFctVal, "__call__"):
                raise (PE.PyAValError("`extractFctVal` needs to be callable!",
                                      solution="Specify a function here or try to use None."))
            self._extractFctVal = extractFctVal
        # Set up ranges
        rs = []
        for par in pars:
            r = ranges[par]
            if len(r) > 4:
                # Use the axis as given
                rs.append(r)
                continue
            if len(r) < 4:
                # By default, use linear spacing
                mode = 'lin'
            else:
                if not isinstance(r[3], six.string_types):
                    raise (PE.PyAValError("If the range has 4 entries, the fourth must be a string specifying the mode.",
                                          solution="Use either 'lin' or 'log' as the fourth entry."))
                mode = r[3]
            if mode == 'lin':
                rs.append(numpy.linspace(r[0], r[1], r[2]))
            elif mode == 'log':
                # Calculate factor
                s = numpy.power((r[1]/r[0]), 1.0/r[2])
                rs.append(r[0] * numpy.power(s, numpy.arange(r[2])))
            else:
                raise (PE.PyAValError("Unknown mode: " + str(mode),
                                      solution="Use either 'lin' or 'log'."))
        # Save state of object
        saveObj = self.saveState()
        saveFitResult = self.fitResult
        saveModels = {}
        for k in six.iterkeys(self._compos):
            saveModels[k] = self.models[k].copy()
        # Freeze parameters, which are affected
        self.freeze(pars)
        # Store result
        result = []
        # Loop over the axes
        nli = pyaC.NestedLoop(list(map(len, rs)))
        for index in nli:
            for i, p in enumerate(pars):
                self[p] = rs[i][index[i]]
            # Fit using previous setting
            # Note that mAA is dispensable, because self.minAlgo will be a callable.
            self.fit(None, None, minAlgo=self.minAlgo, miniFunc=self.miniFunc,
                     *self.fminPars, **self.fminArgs)
            # Build up result
            ppr = []
            for par in pars:
                ppr.append(self[par])
            try:
                ppr.append(self._extractFctVal(self.fitResult))
            except Exception as e:
                PE.warn(PE.PyAValError("The call to the `extractFctVal` function failed. Using full output." +
                                       "\n  Original message: " + str(e)))
                ppr.append(self.fitResult)
            if not quiet:
                print("Result from last iteration:")
                print("  ", ppr)
            ppr.append(index)
            result.append(ppr)
        # Restore old state of object
        self.restoreState(saveObj)
        self.fitResult = saveFitResult
        for k in six.iterkeys(self._compos):
            self.models[k] = saveModels[k]
        return result

    def __init__(self):
        r"""
        Simultaneous model fitting.

        As an example, take a simultaneous measurement of a photometric planetary transit and
        the Rossiter-McLaughlin effect. Surely, both should be described by a subset of common
        parameters like the size of the planet and the large semi-major axis, but the
        models/measurements
        refer to quite different regimes: brightness and radial-velocity shift. This class
        can be used to carry out a fit of both simultaneously. 

        Attributes
        ----------
        pars : Instance of Params
            Manages the model parameters.
        models : dictionary
            A dictionary of the form component-number model; saves the evaluated
            models.
        penaltyFactor : float
            Factor used to scale the penalty imposed if parameter
            restrictions are violated.
        _compos : dictionary
            A dictionary of the form component-number model-component. The
            component number uniquely identifies every model component.

        """
        self._compos = {}
        self.models = {}
        self.pars = None
        self.penaltyFactor = 1e20
        self.naming = ModelNameIdentBase()
        self._stepparEnabled = False



from __future__ import print_function, division
import unittest
import os


class TestExampleSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_firstExample(self):
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Before we can start fitting, we need something to fit.
        # So let us create some data...

        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A": -5.0, "sig": 10.0, "mu": 10.0, "off": 1.0, "lin": 0.0}
        # Calculate profile
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2 * pi * gPar["sig"] ** 2) * exp(
            -((x - gPar["mu"]) ** 2) / (2 * gPar["sig"] ** 2)
        )
        # Add some noise
        y += random.normal(0.0, 0.01, x.size)
        # Let us see what we have done...
        plt.plot(x, y, "bp")

        # Now we can start exploiting the funcFit functionality to
        # fit a Gaussian to our data. In the following lines, we
        # create a fitting object representing a Gaussian and set guess parameters.

        # Now let us come to the fitting
        # First, we create the Gauss1d fit object
        gf = fuf.GaussFit1d()
        # See what parameters are available
        print("List of available parameters: ", gf.availableParameters())
        # Set guess values for the parameters
        gf["A"] = -10.0
        gf["sig"] = 15.77
        gf["off"] = 0.87
        gf["mu"] = 7.5
        # Let us see whether the assignment worked
        print("Parameters and guess values: ")
        print("  A   : ", gf["A"])
        print("  sig : ", gf["sig"])
        print("  off : ", gf["off"])
        print("  mu  : ", gf["mu"])
        print("")

        # Now some of the strengths of funcFit are demonstrated; namely, the
        # ability to consider some parameters as free and others as fixed.
        # By default, all parameters of the GaussFit1d are frozen.

        # Show values and names of frozen parameters
        print("Names and values of FROZEN parameters: ", gf.frozenParameters())

        # Which parameters shall be variable during the fit?
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["A", "sig", "off", "mu"])

        # Let us assume that we know that the amplitude is negative, i.e.,
        # no lower boundary (None) and 0.0 as upper limit.
        gf.setRestriction({"A": [None, 0.0]})

        # Now start the fit
        gf.fit(x, y, yerr=ones(x.size) * 0.01)

        # Write the result to the screen and plot the best fit model
        gf.parameterSummary()
        plt.plot(x, gf.model, "r--")

        # Show the data and the best fit model

    #     plt.show()

    def testsanity_fmin_powell_example(self):
        r"""
        Checking fmin_powell example
        """
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A": -5.0, "sig": 10.0, "mu": 10.0, "off": 1.0, "lin": 0.0}
        # Calculate profile
        x = arange(100) - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2 * pi * gPar["sig"] ** 2) * exp(
            -((x - gPar["mu"]) ** 2) / (2 * gPar["sig"] ** 2)
        )
        # Add some noise
        y += random.normal(0.0, 0.01, x.size)
        # Let us see what we have done...
        plt.plot(x, y, "bp")

        gf = fuf.GaussFit1d()
        # Set guess values for the parameters
        gf["A"] = -1.0
        gf["sig"] = 12.77
        gf["off"] = 1.87
        gf["mu"] = 10.5

        # Which parameters shall be variable during the fit?
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["A", "sig", "off", "mu"])

        # Fit using the algorithm scipy.optimize.fmin_powell
        gf.fit(x, y, yerr=ones(x.size) * 0.01, minAlgo="spfmp")

        # Write the result to the screen and plot the best fit model
        gf.parameterSummary()
        plt.plot(x, gf.model, "r--")

        # Show the data and the best fit model
        # plt.show()

    def testsanity_CustomModel(self):
        # Import numpy and matplotlib
        from numpy import arange, random
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        class StraightLine(fuf.OneDFit):
            r"""
            Implements a straight line of the form y = "off" + x * "lin".
            """

            def __init__(self):
                fuf.OneDFit.__init__(self, ["off", "lin"])

            def evaluate(self, x):
                r"""
the \
lues.

ters:
 at \
odel.
        """
                y = self["off"] + (self["lin"] * x)
                return y

        # Generate some data and add noise
        x = arange(100)
        y = 10.0 + 2.0 * x + random.normal(0.0, 5.0, 100)

        # Create fitting class instance and set initial guess
        # Note that all parameters are frozen by default
        lf = StraightLine()
        lf["off"] = 20.0
        lf["lin"] = 1.0
        # Thaw parameters
        lf.thaw(["off", "lin"])

        # Start fitting
        lf.fit(x, y)

        # Investigate the result
        lf.parameterSummary()
        plt.plot(x, y, "bp")
        plt.plot(x, lf.model, "r--")

    #     plt.show()

    def testsanity_Relations(self):
        # import numpy and matplotlib
        from numpy import arange, random
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        class StraightLine(fuf.OneDFit):
            r"""
            Implements a straight line of the form y = "off" + x * "lin".
            """

            def __init__(self):
                fuf.OneDFit.__init__(self, ["off", "lin"])

            def evaluate(self, x):
                r"""
                Calculates and returns model according to the current parameter values.

                Parameters:
                  - x - Array specifying the positions at which to evaluate the model.
                """
                y = self["off"] + (self["lin"] * x)
                return y

        # Create a function, which defines the relation.

        def getLinearRelation(factor):
            def linOffRel(off):
                r"""
                Function used to relate parameters "lin" and "off".
                """
                return factor * off

            return linOffRel

        # Note, above we used a nested function (a closure) to define
        # the relation. This approach is very flexible. If we were already
        # sure about the value of ``factor'' (e.g., 10.0), we could
        # simply have used:
        #
        # def linOffRel(off):
        #   return 10.0 * off

        # Generate some data with noise
        x = arange(100)
        y = 100.0 + 2.0 * x + random.normal(0.0, 5.0, 100)

        # Create fitting class instance and set initial guess
        lf = StraightLine()
        lf["off"] = 20.0
        lf["lin"] = 1.0
        # Thaw parameters
        lf.thaw(["off", "lin"])

        # Assume we know about a relation between 'lin' and 'off'
        # In particular, lin = 9.0 * off. We use the function getLinearRelation
        # to obtain a function object defining the relation.
        lf.relate("lin", ["off"], getLinearRelation(9))

        # Start fitting
        lf.fit(x, y)

        # Investigate the result
        lf.parameterSummary()
        plt.plot(x, y, "bp")
        plt.plot(x, lf.model, "r--")

    #     plt.show()

    def testsanity_CombiningModels(self):
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Creating Gaussians with some noise
        # Choose some parameters...
        gPar1 = {"A": -5.0, "sig": 10.0, "mu": 20.0, "off": 1.0, "lin": 0.0}
        gPar2 = {"A": +10.0, "sig": 10.0, "mu": -20.0, "off": 0.0, "lin": 0.0}
        # Calculate profile
        x = arange(100) - 50.0
        y = gPar1["off"] + gPar1["A"] / sqrt(2 * pi * gPar1["sig"] ** 2) * exp(
            -((x - gPar1["mu"]) ** 2) / (2 * gPar1["sig"] ** 2)
        )
        y -= gPar2["off"] + gPar2["A"] / sqrt(2 * pi * gPar2["sig"] ** 2) * exp(
            -((x - gPar2["mu"]) ** 2) / (2 * gPar2["sig"] ** 2)
        )
        # Add some noise
        y += random.normal(0.0, 0.01, x.size)
        # Let us see what we have done...
        plt.plot(x, y, "bp")

        # Now let us come to the fitting
        # First, we create two Gauss1d fit objects
        gf1 = fuf.GaussFit1d()
        gf2 = fuf.GaussFit1d()

        # Assign guess values for the parameters
        gf1["A"] = -0.3
        gf1["sig"] = 3.0
        gf1["off"] = 0.0
        gf1["mu"] = +5.0

        gf2["A"] = 3.0
        gf2["sig"] = 15.0
        gf2["off"] = 1.0
        gf2["mu"] = -10.0

        # Which parameters shall be variable during the fit?
        # 'Thaw' those (the order is irrelevant)
        gf1.thaw(["A", "sig", "mu"])
        gf2.thaw(["sig", "mu", "off"])

        # Our actual model is the sum of both Gaussians
        twoG = gf1 + gf2

        # Show a description of the model depending on the
        # names of the individual components
        print()
        print("Description of the model: ", twoG.description())
        print()

        # Note that now the parameter names changed!
        # Each parameter is now named using the "property"
        # (e.g., 'A' or 'sig') as the first part, the component
        # "root name" (in this case 'Gaussian') and a component
        # number in parenthesis.
        print("New parameter names and values: ")
        twoG.parameterSummary()

        # We forgot to thaw the amplitude of the second Gaussian, but
        # we can still do it, but we have to refer to the correct name:
        # either by using the (new) variable name:
        twoG.thaw("A_Gaussian(2)")
        # or by specifying property name, root name, and component number
        # separately (note that a tuple is used to encapsulate them):
        twoG.thaw(("A", "Gaussian", 2))
        # We decide to rather freeze the offset of the second
        # Gaussian (we could have used a tuple here, too).
        twoG.freeze("off_Gaussian(2)")

        # Start fit as usual
        twoG.fit(x, y, yerr=ones(x.size) * 0.01)

        # Write the result to the screen and plot the best fit model
        print()
        print("--------------------------------")
        print("Parameters for the combined fit:")
        print("--------------------------------")
        twoG.parameterSummary()

        # Show the data and the best fit model
        plt.plot(x, twoG.model, "r--")

    #     plt.show()

    def testsanity_CustomObjectiveFunctions(self):
        # Import numpy and matplotlib
        from numpy import arange, exp, random, ones, sum, abs
        import matplotlib.pylab as plt

        # Import funcFit
        from PyAstronomy import funcFit as fuf

        # Define parameters of faked data
        A = 1.0
        tau = 10.0
        off = 0.2
        t0 = 40.0

        # Calculate fake data set
        x = arange(100)
        y = A * exp(-(x - t0) / tau) * (x > t0) + off
        y += random.normal(0.0, 0.1, 100)
        yerr = ones(100) * 0.01

        # Exponential decay model
        edf = fuf.ExpDecayFit1d()

        # Define free quantities
        edf.thaw(["A", "tau", "off", "t0"])
        # Let the amplitude be positive
        edf.setRestriction({"A": [0.0, None]})
        # Define initial guess
        edf.assignValue({"A": 1.0, "tau": 15.0, "off": 0.2, "t0": 50.0})

        # Do not use chi square, but the linear deviation from model
        # to evaluate quality of fit.
        # Use the "MiniFunc" decorator to define your custom objective
        # function. This decorator takes the fitting object as an
        # argument. The function has to accept two arguments: the
        # fitting object and the list of free parameters.
        @fuf.MiniFunc(edf)
        def mini(edf, P):
            m = sum(abs(edf.model - edf.y) / edf.yerr)
            print("mini - current parameters: ", P, ", value is: ", m)
            return m

        # Carry out fit WITH SELF-DEFINED OBJECTIVE FUNCTION
        edf.fit(x, y, yerr=yerr, miniFunc=mini)

        # Show parameter values and plot best-fit model.
        edf.parameterSummary()
        plt.errorbar(x, y, yerr)
        plt.plot(x, edf.model, "r-")

    #     plt.show()

    def testsanity_Overbinning(self):
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Creating a Gaussian with some noise
        # Choose some parameters...
        gPar = {"A": -5.0, "sig": 10.0, "mu": 10.0, "off": 1.0, "lin": 0.0}
        # Calculate profile
        x = arange(20) / 20.0 * 100.0 - 50.0
        y = gPar["off"] + gPar["A"] / sqrt(2 * pi * gPar["sig"] ** 2) * exp(
            -((x - gPar["mu"]) ** 2) / (2 * gPar["sig"] ** 2)
        )
        # Add some noise
        y += random.normal(0.0, 0.01, x.size)
        # Let us see what we have done...
        plt.plot(x, y, "bp")

        # First, we create a "GaussFit1d_Rebin" class object (note that the
        # class object has still to be instantiated, the name is arbitrary).
        GaussFit1d_Rebin = fuf.turnIntoRebin(fuf.GaussFit1d)
        # Do the instantiation and specify how the overbinning should be
        # carried out.
        gf = GaussFit1d_Rebin()
        gf.setRebinArray_Ndt(x, 10, x[1] - x[0])
        # See what parameters are available
        print("List of available parameters: ", gf.availableParameters())
        # Set guess values for the parameters
        gf["A"] = -10.0
        gf["sig"] = 15.77
        gf["off"] = 0.87
        gf["mu"] = 7.5
        # Let us see whether the assignment worked
        print("Parameters and guess values: ")
        print("  A   : ", gf["A"])
        print("  sig : ", gf["sig"])
        print("  off : ", gf["off"])
        print("  mu  : ", gf["mu"])
        print("")

        # Now some of the strengths of funcFit are demonstrated; namely, the
        # ability to consider some parameters as free and others as fixed.
        # By default, all parameters of the GaussFit1d are frozen.

        # Show values and names of frozen parameters
        print("Names and values if FROZEN parameters: ", gf.frozenParameters())

        # Which parameters shall be variable during the fit?
        # 'Thaw' those (the order is irrelevant)
        gf.thaw(["A", "sig", "off", "mu"])

        # Let us assume that we know that the amplitude is negative, i.e.,
        # no lower boundary (None) and 0.0 as upper limit.
        gf.setRestriction({"A": [None, 0.0]})

        # Now start the fit
        gf.fit(x, y, yerr=ones(x.size) * 0.01)

        # Write the result to the screen and plot the best fit model
        gf.parameterSummary()
        # Plot the final best-fit model
        plt.plot(x, gf.model, "rp--")
        # Show the overbinned (=unbinned) model, indicate by color
        # which point are averaged to obtain a point in the binned
        # model.
        for k, v in gf.rebinIdent.items():
            c = "y"
            if k % 2 == 0:
                c = "k"
            plt.plot(gf.rebinTimes[v], gf.unbinnedModel[v], c + ".")

        # Show the data and the best fit model

    #     plt.show()

    def testsanity_simultaneousFit(self):
        from PyAstronomy import funcFit as fuf
        import numpy
        import matplotlib.pylab as plt

        # Set up two different x axes.
        x1 = numpy.arange(100.0) / 100.0 - 0.5
        x2 = numpy.arange(150.0) / 150.0 - 0.25

        # Getting the models ...
        gauss = fuf.GaussFit1d()
        calor = fuf.CauchyLorentz1d()
        # and assign parameters.
        gauss.assignValue(
            {"A": 0.02, "sig": 0.1, "mu": 0.0, "off": 1.0, "lin": 0.0})
        calor.assignValue(
            {"A": 0.07, "g": 0.1, "mu": 0.2, "off": 1.0, "lin": 0.0})

        # Create noisy data.
        y1 = gauss.evaluate(x1) + numpy.random.normal(0.0, 0.01, 100)
        y2 = calor.evaluate(x2) + numpy.random.normal(0.0, 0.01, 150)

        # Plot the noisy data.
        plt.subplot(2, 1, 1)
        plt.errorbar(x1, y1, yerr=numpy.ones(100) * 0.01)
        plt.subplot(2, 1, 2)
        plt.errorbar(x2, y2, yerr=numpy.ones(150) * 0.01)

        # Now, get ready two fit the data sets simultaneously.
        sf = fuf.SyncFitContainer()
        # Tell the class about the two components and save the
        # component numbers assigned to them:
        gaussCno = sf.addComponent(gauss)
        calorCno = sf.addComponent(calor)

        print("Component numbers in the syncFit container:")
        print("  Gauss: ", gaussCno, ",  Cauchy-Lorentz: ", calorCno)
        print()

        # See what happened to the parameters in the
        # simultaneous fitting class.
        # The variable names have changed.
        sf.parameterSummary()

        # Thaw all parameters (for later fit) ...
        sf.thaw(list(sf.parameters()))
        # but not the linear term.
        sf.freeze(["lin_Gaussian[s1]", "lin_CauLor[s2]"])

        # Tell the class about the identity of parameters,
        # either by using the "property name" of the parameter:
        sf.treatAsEqual("off")
        # or by specifying the names explicitly.
        sf.treatAsEqual(["g_CauLor[s2]", "sig_Gaussian[s1]"])

        # See what happened to the parameters in the
        # simultaneous fitting class.
        print()
        print("Parameters after 'treatAsEqual' has been applied:")
        sf.parameterSummary()

        # Randomize starting values.
        for fp in sf.freeParamNames():
            sf[fp] = sf[fp] + numpy.random.normal(0.0, 0.05)

        # Set up the data appropriately.
        data = {gaussCno: [x1, y1], calorCno: [x2, y2]}
        yerr = {gaussCno: numpy.ones(
            100) * 0.01, calorCno: numpy.ones(150) * 0.01}

        # Start the fit.
        sf.fit(data, yerr=yerr)

        # Show the best-fit values.
        print()
        print("Best-fit parameters:")
        sf.parameterSummary()

        # Plot the best-fit model(s).
        plt.subplot(2, 1, 1)
        plt.plot(x1, sf.models[gaussCno], "r--")
        plt.subplot(2, 1, 2)
        plt.plot(x2, sf.models[calorCno], "r--")

    #     plt.show()

    def testsanity_2dCircularFit(self):
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Get the circular model and assign
        # parameter values
        c = fuf.Circle2d()
        c["r"] = 1.0
        c["t0"] = 0.0
        c["per"] = 3.0

        # Evaluate the model at a number of
        # time stamps
        t = np.linspace(0.0, 10.0, 20)
        pos = c.evaluate(t)

        # Add some error to the "measurement"
        pos += np.reshape(np.random.normal(0.0, 0.2, pos.size), pos.shape)
        err = np.reshape(np.ones(pos.size), pos.shape) * 0.2

        # Define free parameters and fit the model
        c.thaw(["r", "t0", "per"])
        c.fit(t, pos, yerr=err)
        c.parameterSummary()

        # Evaluate the model at a larger number of
        # points for plotting
        tt = np.linspace(0.0, 10.0, 200)
        model = c.evaluate(tt)

        # Plot the result
        plt.errorbar(pos[::, 0], pos[::, 1], yerr=err[::, 1],
                     xerr=err[::, 0], fmt="bp")
        plt.plot(model[::, 0], model[::, 1], "r--")

    #    plt.show()

    def testsanity_2dGaussFit(self):
        from PyAstronomy import funcFit as fuf
        import numpy as np
        import matplotlib.pylab as plt

        # Constructing the individual coordinate axes
        x = np.linspace(-2.0, 2.0, 50)
        y = np.linspace(-2.0, 2.0, 50)
        # Applying funcFit's "coordinateGrid" helper function
        # to built appropriate array-index -> coordinate mapping
        # needed for nD fitting.
        g = fuf.coordinateGrid(x, y)

        # Create the 2d-Gaussian model and assign
        # some model parameters.
        gf = fuf.GaussFit2d()
        gf["sigx"] = 0.75
        gf["sigy"] = 0.4
        gf["A"] = 1.0
        gf["rho"] = 0.4

        # Get the "data" by evaluating the model
        # and adding some noise. Note that the coordinate
        # mapping (array g) is passed to evaluate here.
        im = gf.evaluate(g)
        im += np.reshape(np.random.normal(0.0, 0.1, 2500), (50, 50))
        err = np.ones((50, 50)) * 0.1

        # Thaw parameters and fit
        gf.thaw(["A", "rho"])
        gf.fit(g, im, yerr=err)

        # Show the resulting parameter values ...
        gf.parameterSummary()

        # ... and plot the result.
        plt.title("Image data")
        plt.imshow(np.transpose(im), origin="lower")
        #    plt.show()
        plt.title("Residuals")
        plt.imshow(np.transpose(im - gf.evaluate(g)), origin="lower")

    #    plt.show()

    def testsanity_2gGaussFitTupleExample(self):
        from PyAstronomy import funcFit as fuf
        import numpy as np
        import matplotlib.pylab as plt

        # Constructing the individual coordinate axes
        x = np.linspace(-2.0, 2.0, 50)
        y = np.linspace(-2.0, 2.0, 50)

        # Create the 2d-Gaussian model and assign
        # some model parameters.
        gf = fuf.GaussFit2dTuple()
        gf["sigx"] = 0.75
        gf["sigy"] = 0.4
        gf["A"] = 1.0
        gf["rho"] = 0.4

        # Get the "data" by evaluating the model
        # and adding some noise. Note that the coordinate
        # mapping (array g) is passed to evaluate here.
        im = gf.evaluate((x, y))
        im += np.reshape(np.random.normal(0.0, 0.1, 2500), (50, 50))
        err = np.ones((50, 50)) * 0.1

        # Thaw parameters and fit
        gf.thaw(["A", "rho"])
        gf.fit((x, y), im, yerr=err)

        # Show the resulting parameter values ...
        gf.parameterSummary()

        # ... and plot the result.
        plt.title("Image data")
        plt.imshow(np.transpose(im), origin="lower")
        #    plt.show()
        plt.title("Residuals")
        plt.imshow(np.transpose(im - gf.evaluate((x, y))), origin="lower")

    #    plt.show()

    def testsanity_coordinateGridExample(self):
        from PyAstronomy import funcFit as fuf
        import numpy as np

        # Constructing the two individual coordinate axes
        x = np.linspace(-2.0, 2.0, 50)
        y = np.linspace(-2.0, 2.0, 50)

        # Applying funcFit's "coordinateGrid" helper function
        # to built appropriate array-index -> coordinate mapping
        # needed for nD fitting.
        g = fuf.coordinateGrid(x, y)

        print("(x, y) coordinates at index (11, 28): ", g[11, 28])

    def testsanity_CashStatisticsExample(self):
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Get a Gaussian fitting object and
        # set some parameters
        g = fuf.GaussFit1d()
        g["A"] = 5.1
        g["sig"] = 0.5
        g["mu"] = 3.94

        # Generate some data with Poisson statistics
        x = np.linspace(0.0, 7.0, 50)
        y = np.zeros(len(x))
        for i in range(len(x)):
            y[i] = np.random.poisson(g.evaluate(x[i]))

        # Choose free parameters and "disturb" the
        # starting parameters for the fit a little.
        g.thaw(["A", "sig", "mu"])
        for par in g.freeParamNames():
            g[par] += np.random.normal(0.0, g[par] * 0.1)

        # Fit using Cash statistic and print out
        # result.
        g.fit(x, y, miniFunc="cash79")
        g.parameterSummary()

        # Plot the result
        plt.plot(x, y, "bp")
        plt.plot(x, g.evaluate(x), "r--")

    #    plt.show()

    def testsanity_steppar1(self):
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Set up a Gaussian model
        # and create some "data"
        x = np.linspace(0, 2, 100)
        gf = fuf.GaussFit1d()
        gf["A"] = 0.87
        gf["mu"] = 1.0
        gf["sig"] = 0.2
        y = gf.evaluate(x)
        y += np.random.normal(0.0, 0.1, len(x))

        # Thaw parameters, which are to be fitted. Note
        # that those parameters will also be fitted during
        # the stepping; no further parameters will be thawed.
        gf.thaw(["A", "mu", "sig"])
        # ... and "disturb" starting values a little.
        gf["A"] = gf["A"] + np.random.normal(0.0, 0.1)
        gf["mu"] = gf["mu"] + np.random.normal(0.0, 0.1)
        gf["sig"] = gf["sig"] + np.random.normal(0.0, 0.03)
        # Find the best fit solution
        gf.fit(x, y, yerr=np.ones(len(x)) * 0.1)

        # Step the amplitude (area of the Gaussian) through
        # the range 0.8 to 0.95 in 20 steps. Note that the
        # last part of `ranges` ('lin') is optional. You may
        # also use `log`; in this case, the stepping would be
        # equidistant in the logarithm.
        # In each step of `A`, "mu" and "sig" will be fitted,
        # because they had been thawed earlier.
        sp = gf.steppar("A", ranges={"A": [0.8, 0.95, 20, "lin"]})
        # Extract the values for the Gaussian normalization
        # (amplitude) ...
        As = list(map(lambda x: x[0], sp))
        # ... and chi square.
        chis = list(map(lambda x: x[1], sp))

        # Find minimum chi square
        cmin = min(chis)

        # Plot A vs. chi square
        plt.title(r"A vs. $\chi^2$ with 68% and 90% confidence levels")
        plt.xlabel("A")
        plt.ylabel(r"$\chi^2$")
        plt.plot(As, chis, "bp-")
        plt.plot(As, [cmin + 1.0] * len(As), "k--")
        plt.plot(As, [cmin + 2.706] * len(As), "k:")

    #    plt.show()

    def testsanity_steppar2(self):
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Set up a Gaussian model
        # and create some "data"
        x = np.linspace(0, 2, 100)
        gf = fuf.GaussFit1d()
        gf["A"] = 0.87
        gf["mu"] = 1.0
        gf["sig"] = 0.2
        y = gf.evaluate(x)
        y += np.random.normal(0.0, 0.1, len(x))

        # Thaw parameters, which are to be fitted ...
        gf.thaw(["A", "mu", "sig"])
        # ... and "disturb" starting values a little.
        gf["A"] = gf["A"] + np.random.normal(0.0, 0.1)
        gf["mu"] = gf["mu"] + np.random.normal(0.0, 0.1)
        gf["sig"] = gf["sig"] + np.random.normal(0.0, 0.03)
        # Find the best fit solution
        gf.fit(x, y, yerr=np.ones(len(x)) * 0.1)

        # Step the amplitude (area of the Gaussian) and the
        # center ("mu") of the Gaussian through the given
        # ranges.
        sp = gf.steppar(
            ["A", "mu"], ranges={"A": [0.8, 0.95, 20], "mu": [0.96, 1.05, 15]}
        )

        # Get the values for `A`, `mu`, and chi-square
        # from the output of steppar.
        As = list(map(lambda x: x[0], sp))
        mus = list(map(lambda x: x[1], sp))
        chis = list(map(lambda x: x[2], sp))

        # Create a chi-square array using the
        # indices contained in the output.
        z = np.zeros((20, 15))
        for s in sp:
            z[s[3]] = s[2]

        # Find minimum chi-square and define levels
        # for 68%, 90%, and 99% confidence intervals.
        cm = min(chis)
        levels = [cm + 2.3, cm + 4.61, cm + 9.21]

        # Plot the contours to explore the confidence
        # interval and correlation.
        plt.xlabel("mu")
        plt.ylabel("A")
        plt.contour(np.sort(np.unique(mus)), np.sort(
            np.unique(As)), z, levels=levels)
        # Plot the input value
        plt.plot([1.0], [0.87], "k+", markersize=20)

    #    plt.show()

    def testsanity_errorConfInterval(self):
        r"""
        Checking example of errorConfInterval
        """
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Set up a Gaussian model
        # and create some "data"
        x = np.linspace(0, 2, 100)
        gf = fuf.GaussFit1d()
        gf["A"] = 0.87
        gf["mu"] = 1.0
        gf["sig"] = 0.2
        y = gf.evaluate(x)
        y += np.random.normal(0.0, 0.1, len(x))

        # Thaw parameters, which are to be fitted. Note
        # that those parameters will also be fitted during
        # the stepping; no further parameters will be thawed.
        gf.thaw(["A", "mu", "sig"])
        # ... and "disturb" starting values a little.
        gf["A"] = gf["A"] + np.random.normal(0.0, 0.1)
        gf["mu"] = gf["mu"] + np.random.normal(0.0, 0.1)
        gf["sig"] = gf["sig"] + np.random.normal(0.0, 0.03)
        # Find the best fit solution
        gf.fit(x, y, yerr=np.ones(len(x)) * 0.1)

        # Step the amplitude (area of the Gaussian) through
        # the range 0.8 to 0.95 in 20 steps. Note that the
        # last part of `ranges` ('lin') is optional. You may
        # also use `log`; in this case, the stepping would be
        # equidistant in the logarithm.
        # In each step of `A`, "mu" and "sig" will be fitted,
        # because they had been thawed earlier.
        sp = gf.steppar("A", ranges={"A": [0.8, 0.95, 20, "lin"]})
        # Extract the values for the Gaussian normalization
        # (amplitude) ...
        As = [x[0] for x in sp]
        # ... and chi square.
        chis = [x[1] for x in sp]

        # Calculate the confidence interval automatically
        cfi90 = gf.errorConfInterval("A", dstat=2.706)
        print("90% Confidence interval: ", cfi90["limits"])
        print("  corresponding objective function values: ", cfi90["OFVals"])
        print("  number of iterations needed: ", cfi90["iters"])

        cfi68 = gf.errorConfInterval("A", dstat=1.0)
        print("68% Confidence interval: ", cfi68["limits"])
        print("  corresponding objective function values: ", cfi68["OFVals"])
        print("  number of iterations needed: ", cfi68["iters"])

        # Plot A vs. chi square
        plt.title(r"A vs. $\chi^2$ 90% (black) and 68% (blue) confidence intervals")
        plt.xlabel("A")
        plt.ylabel(r"$\chi^2$")
        plt.plot(As, chis, "bp-")
        # Indicate confidence levels by vertical lines
        plt.plot(As, [cfi90["OFMin"] + 1.0] * len(As), "g:")
        plt.plot(As, [cfi90["OFMin"] + 2.706] * len(As), "g:")
        # PLot lines to indicate confidence intervals
        plt.plot([cfi90["limits"][0]] * 2, [min(chis), max(chis)], "k--")
        plt.plot([cfi90["limits"][1]] * 2, [min(chis), max(chis)], "k--")
        plt.plot([cfi68["limits"][0]] * 2, [min(chis), max(chis)], "b--")
        plt.plot([cfi68["limits"][1]] * 2, [min(chis), max(chis)], "b--")

    #     plt.show()

    def testsanity_conditionalRestrictions(self):
        r"""
        Check the conditional restriction example.
        """
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import funcFit as fuf

        # Get fitting object for a Gaussian ...
        g = fuf.GaussFit1d()
        # .. and define the parameters
        g["A"] = 0.97
        g["mu"] = 0.1
        g["sig"] = 0.06

        # Generate some "data" with noise included
        x = np.linspace(-1.0, 1.0, 200)
        y = g.evaluate(x) + np.random.normal(0.0, 0.1, len(x))
        yerr = np.ones(len(x)) * 0.1

        def myRestriction(A, sig):
            r"""
            A conditional restriction.

            Returns
            -------
            Penalty : float
                A large value if condition is violated
                and zero otherwise.
            """
            if A > 10.0 * sig:
                return np.abs(A - 10.0 * sig + 1.0) * 1e20
            return 0.0

        # Add the conditional restriction to the model and save
        # the unique ID, which can be used to refer to that
        # restriction.
        uid = g.addConditionalRestriction(["A", "sig"], myRestriction)
        print("Conditional restriction has been assigned the ID: ", uid)
        print()

        # Now see whether the restriction is really in place
        g.showConditionalRestrictions()

        # Define free parameters ...
        g.thaw(["A", "mu", "sig"])
        # ... and fit the model (restriction included)
        g.fit(x, y, yerr=yerr)

        # Save the resulting best-fit model
        restrictedModel = g.model.copy()

        # Remove the conditional restriction and re-fit
        g.removeConditionalRestriction(uid)
        g.fit(x, y, yerr=yerr)

        # Save new model
        unrestrictedModel = g.model.copy()

        # Plot the result


#     plt.errorbar(x, y, yerr=yerr, fmt='b.')
#     plt.plot(x, restrictedModel, 'r--', label="Restricted")
#     plt.plot(x, unrestrictedModel, 'g--', label="Unrestricted")
#     plt.legend()
#     plt.show()


class TestMCMCExampleSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        try:
            os.remove("mcmcExample.tmp")
        except:
            print("Could not remove file: mcmcExample.tmp")
        try:
            os.remove("mcmcTA.emcee")
        except:
            print("Could not remove file: mcmcTA.emcee")
        try:
            os.remove("mcmcSample.tmp")
        except:
            print("Could not remove file: mcmcSample.tmp")
        try:
            os.remove("chain.emcee")
        except:
            pass
        try:
            os.remove("gauss.emcee")
        except:
            print("Could not remove file: gauss.emcee")
        try:
            os.remove("musig.emcee")
        except:
            print("Could not remove file: musig.emcee")

    # ===============================================================================
    #   def testsanity_MCMCSampler(self):
    #     # Import some required modules
    #     from numpy import arange, sqrt, exp, pi, random, ones
    #     import matplotlib.pylab as plt
    #     import pymc
    #     # ... and now the funcFit package
    #     from PyAstronomy import funcFit as fuf
    #
    #     # Creating a Gaussian with some noise
    #     # Choose some parameters...
    #     gPar = {"A":-5.0, "sig":10.0, "mu":10.0, "off":1.0, "lin":0.0}
    #     # Calculate profile
    #     x = arange(100) - 50.0
    #     y = gPar["off"] + gPar["A"] / sqrt(2*pi*gPar["sig"]**2) \
    #     * exp(-(x-gPar["mu"])**2/(2*gPar["sig"]**2))
    #     # Add some noise
    #     y += random.normal(0.0, 0.01, x.size)
    #
    #     # Now let us come to the fitting
    #     # First, we create the Gauss1d fit object
    #     gf = fuf.GaussFit1d()
    #     # See what parameters are available
    #     print("List of available parameters: ", gf.availableParameters())
    #     # Set guess values for the parameters
    #     gf["A"] = -10.0
    #     gf["sig"] = 15.77
    #     gf["off"] = 0.87
    #     gf["mu"] = 7.5
    #     # Let us see whether the assignment worked
    #     print("Parameters and guess values: ", gf.parameters())
    #
    #     # Which parameters shall be variable during the fit?
    #     # 'Thaw' those (the order is irrelevant)
    #     gf.thaw(["A", "sig", "off", "mu"])
    #
    #     # Now start a simplex fit
    #     gf.fit(x,y,yerr=ones(x.size)*0.01)
    #
    #     # Obtain the best-fit values derived by the simplex fit.
    #     # They are to be used as start values for the MCMC sampling.
    #     # Note that 'A' is missing - we will introduce this later.
    #     X0 = {"sig":gf["sig"], "off":gf["off"], "mu":gf["mu"]}
    #
    #     # Now we specify the limits within which the individual parameters
    #     # can be varied (for those parameters listed in the 'X0' dictionary).
    #     Lims = {"sig":[-20.,20.], "off":[0.,2.], "mu":[5.,15.]}
    #
    #     # For the parameters contained in 'X0', define the step widths, which
    #     # are to be used by the MCMC sampler. The steps are specified using
    #     # the same scale/units as the actual parameters.
    #     steps = {"A":0.01, "sig":0.1, "off":0.1, "mu":0.1}
    #
    #     # In this example, we wish to define our ``own'' PyMC variable for the parameter
    #     # 'A'. This can be useful, if nonstandard behavior is desired. Note that this
    #     # is an optional parameter and you could simply include the parameter 'A' into
    #     # The framework of X0, Lims, and steps.
    #     ppa = {}
    #     ppa["A"] = pymc.Uniform("A", value=gf["A"], lower=-20., \
    #                             upper=10.0, doc="Amplitude")
    #
    #     # Start the sampling. The resulting Marchov-Chain will be written
    #     # to the file 'mcmcExample.tmp'. In default configuration, pickle
    #     # is used to write that file.
    #     # To save the chain to a compressed 'hdf5'
    #     # file, you have to specify the dbArgs keyword; e.g., use:
    #     #   dbArgs = {"db":"hdf5", "dbname":"mcmcExample.hdf5"}
    #     gf.fitMCMC(x, y, X0, Lims, steps, yerr=ones(x.size)*0.01, \
    #                 pymcPars=ppa, iter=2500, burn=0, thin=1, \
    #                 dbfile="mcmcExample.tmp")
    #
    #     # Reload the database (here, this is actually not required, but it is
    #     # if the Marchov chain is to be analyzed later).
    #     db = pymc.database.pickle.load('mcmcExample.tmp')
    #     # Plot the trace of the amplitude, 'A'.
    #     plt.hist(db.trace("A", 0)[:])
    # #     plt.show()
    # ===============================================================================

    # =============================================================================
    # def testsanity_MCMCPriorExample(self):
    #   from PyAstronomy import funcFit as fuf
    #   import numpy as np
    #   import matplotlib.pylab as plt
    #   import pymc
    #
    #   # Create a Gauss-fit object
    #   gf = fuf.GaussFit1d()
    #
    #   # Choose some parameters
    #   gf["A"] = -0.65
    #   gf["mu"] = 1.0
    #   gf["lin"] = 0.0
    #   gf["off"] = 1.1
    #   gf["sig"] = 0.2
    #
    #   # Simulate data with noise
    #   x = np.linspace(0., 2., 100)
    #   y = gf.evaluate(x)
    #   y += np.random.normal(0, 0.05, len(x))
    #
    #   gf.thaw(["A", "off", "mu", "sig"])
    #
    #   # Set up a normal prior for the offset parameter
    #   # Note!---The name (first parameter) must correspond to that
    #   #         of the parameter.
    #   # The expectation value us set to 0.9 while the width is given
    #   # as 0.01 (tau = 1/sigma**2). The starting value is specified
    #   # as 1.0.
    #   offPar = pymc.Normal("off", mu=0.9, tau=(1./0.01)**2, value=1.0)
    #   # Use a uniform prior for mu.
    #   muPar = pymc.Uniform("mu", lower=0.95, upper=0.97, value=0.96)
    #
    #   # Collect the "extra"-variables in a dictionary using
    #   # their names as keys
    #   pymcPars = {"mu":muPar, "off":offPar}
    #
    #   # Specify starting values, X0, and limits, lims, for
    #   # those parameter distributions not given specifically.
    #   X0 = {"A":gf["A"], "sig":gf["sig"]}
    #   lims = {"A":[-1.0,0.0], "sig":[0., 1.0]}
    #   # Still, the steps dictionary has to contain all
    #   # parameter distributions.
    #   steps = {"A":0.02, "sig":0.02, "mu":0.01, "off":0.01}
    #
    #   # Carry out the MCMC sampling
    #   gf.fitMCMC(x, y, X0, lims, steps, yerr=np.ones(len(x))*0.05, \
    #              pymcPars=pymcPars, burn=1000, iter=3000)
    #
    #   # Setting parameters to mean values
    #   for p in gf.freeParameters():
    #     gf[p] = gf.MCMC.trace(p)[:].mean()
    #
    #   # Show the "data" and model in the upper panel
    #   plt.subplot(2,1,1)
    #   plt.title("Data and model")
    #   plt.errorbar(x, y, yerr=np.ones(len(x))*0.05, fmt="bp")
    #   # Plot lowest deviance solution
    #   plt.plot(x, gf.evaluate(x), 'r--')
    #
    #   # Show the residuals in the lower panel
    #   plt.subplot(2,1,2)
    #   plt.title("Residuals")
    #   plt.errorbar(x, y-gf.evaluate(x), yerr=np.ones(len(x))*0.05, fmt="bp")
    #   plt.plot([min(x), max(x)], [0.0,0.0], 'r-')
    #
    #   #plt.show()
    # =============================================================================

    # ===============================================================================
    #   def testsanity_autoMCMCExample1(self):
    #     from PyAstronomy import funcFit as fuf
    #     import numpy as np
    #     import matplotlib.pylab as plt
    #
    #     x = np.linspace(0,30,1000)
    #     gauss = fuf.GaussFit1d()
    #     gauss["A"] = 1
    #     gauss["mu"] = 23.
    #     gauss["sig"] = 0.5
    #     # Generate some "data" to fit
    #     yerr = np.random.normal(0., 0.05, len(x))
    #     y = gauss.evaluate(x) + yerr
    #     # Thaw the parameters A, mu, and sig
    #     gauss.thaw(["A","mu","sig"])
    #
    #     # Define the ranges, which are used to construct the
    #     # uniform priors and step sizes.
    #     # Note that for "sig", we give only a single value.
    #     # In this case, the limits for the uniform prior will
    #     # be constructed as [m0-1.5, m0+1.5], where m0 is the
    #     # starting value interpreted as the current value of
    #     # mu (23. in this case).
    #     ranges = {"A":[0,10],"mu":3, "sig":[0.1,1.0]}
    #     # Generate default input for X0, lims, and steps
    #     X0, lims, steps = gauss.MCMCautoParameters(ranges)
    #
    #     # Show what happened...
    #     print()
    #     print("Auto-generated input parameters:")
    #     print("X0: ", X0)
    #     print("lims: ", lims)
    #     print("steps: ", steps)
    #     print()
    #     # Call the usual sampler
    #     gauss.fitMCMC(x, y, X0, lims, steps, yerr=yerr, iter=1000)
    #
    #     # and plot the results
    #     plt.plot(x, y, 'k+')
    #     plt.plot(x, gauss.evaluate(x), 'r--')
    # #     plt.show()
    # ===============================================================================

    # ===============================================================================
    #   def testsanity_autoMCMCExample2(self):
    #     from PyAstronomy import funcFit as fuf
    #     import numpy as np
    #     import matplotlib.pylab as plt
    #
    #     x = np.linspace(0,30,1000)
    #     gauss = fuf.GaussFit1d()
    #     gauss["A"] = 1
    #     gauss["mu"] = 23.
    #     gauss["sig"] = 0.5
    #     # Generate some "data" to fit
    #     yerr = np.random.normal(0., 0.05, len(x))
    #     y = gauss.evaluate(x) + yerr
    #
    #     # Define the ranges, which are used to construct the
    #     # uniform priors and step sizes.
    #     # Note that for "sig", we give only a single value.
    #     # In this case, the limits for the uniform prior will
    #     # be constructed as [m0-1.5, m0+1.5], where m0 is the
    #     # starting value interpreted as the current value of
    #     # mu (23. in this case).
    #     ranges = {"A":[0,10],"mu":3, "sig":[0.1,1.0]}
    #
    #     # Call the auto-sampler
    #     # Note that we set picky to False here. In this case, the
    #     # parameters specified in ranges will be thawed automatically.
    #     # All parameters not mentioned there, will be frozen.
    #     gauss.autoFitMCMC(x, y, ranges, yerr=yerr, picky=False, iter=1000)
    #
    #     # and plot the results
    #     plt.plot(x, y, 'k+')
    #     plt.plot(x, gauss.evaluate(x), 'r--')
    # #    plt.show()
    # ===============================================================================

    def testsanity_TAtut_createTrace(self):
        r"""
        TA tutorial, all examples
        """
        # Import some required modules
        import numpy as np
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Starting from with Voigt profile
        vp = fuf.Voigt1d()
        # Set some values to create a model
        vp["A"] = -0.4
        vp["al"] = 0.7
        vp["mu"] = 5500.0
        vp["ad"] = 0.3
        vp["off"] = 1.0

        x = np.linspace(5490.0, 5510.0, 200)
        # Create our data with some noise
        yerr = np.ones(len(x)) * 0.01
        y = vp.evaluate(x) + np.random.normal(0.0, 0.01, len(x))

        # Say, we have a guess of the parameters, which is, however,
        # not entirely correct
        vp["A"] = -0.376
        vp["al"] = 0.9
        vp["mu"] = 5499.7
        vp["ad"] = 0.4
        vp["off"] = 1.0

        # Plot the data and our guess
        plt.errorbar(x, y, yerr=yerr, fmt="b.-")
        plt.plot(x, vp.evaluate(x), "r--")
        plt.show()

        # Thaw the parameters, which we wish to vary
        # during the sampling
        vp.thaw(["A", "al", "mu", "ad"])

        # Use current parameters as starting point for the sampling
        X0 = vp.freeParameters()
        print("Starting point for sampling: ", X0)

        priors = {
            "al": fuf.FuFPrior("limuniform", lower=0.0, upper=100.0),
            "ad": fuf.FuFPrior("limuniform", lower=0.0, upper=100.0),
        }

        # Note that the filename should end in .emcee. Substitute this filename
        # in the following examples.
        vp.fitEMCEE(
            x,
            y,
            yerr=yerr,
            sampleArgs={"iters": 2500},
            dbfile="mcmcTA.emcee",
            priors=priors,
        )

        # Second example

        from PyAstronomy import funcFit as fuf

        # Create an instance of TraceAnalysis
        # telling it which file to use
        ta = fuf.TraceAnalysis("mcmcTA.emcee")

        # Have a look at the deviance to check if and when
        # the chains reached equilibrium.
        ta.plotTrace("deviance")
        #    ta.show()

        # Say, we are sure that after 500 iterations, the chain
        # reached equilibrium. We use this as the burn-in phase
        ta.setBurn(500)

        # Have a second look at the deviance, this time considering
        # the burn-in. Note that the first 500 iterations are not
        # removed from the chain. They are just not considered any
        # more.
        ta.plotTrace("deviance")
        #    ta.show()

        # Third example

        from PyAstronomy import funcFit as fuf

        # Create an instance of TraceAnalysis
        # telling it which file to use
        ta = fuf.TraceAnalysis("mcmcTA.emcee")

        # Use the burn-in from the previous example
        ta.setBurn(500)

        # See which model parameters have been sampled
        print("Available parameters: ", ta.availableParameters())

        # Access the traces of these parameters
        print("Trace for A: ", ta["A"])

        # Calculate mean, median, standard deviation, and
        # credibility interval for the available parameters
        for p in ta.availableParameters():
            hpd = ta.hpd(p, cred=0.95)
            print(
                "Parameter %5s, mean = % g, median = % g, std = % g, 95%% HPD = % g - % g"
                % (p, ta.mean(p), ta.median(p), ta.std(p), hpd[0], hpd[1])
            )

        # Fourth example

        from PyAstronomy import funcFit as fuf

        # Create an instance of TraceAnalysis
        # telling it which file to use
        ta = fuf.TraceAnalysis("mcmcTA.emcee")

        # Use the burn-in from the previous example
        ta.setBurn(500)

        # Have a look at the parameter correlations
        ta.correlationTable()

        # Calculate Pearson's and Spearman's r-coefficients
        print("Pearson: ", ta.pearsonr("ad", "al"))
        print("Spearman: ", ta.spearmanr("ad", "al"))

        # Show a plot of the correlation
        # Note that the plotCorrEnh method can also
        # be used, which is useful in the case of long
        # chains.
        ta.plotCorr(parsList=["ad", "al"])
        # ta.plotCorrEnh(parsList=["ad", "al"])
        #     ta.show()

        # Fifth example

        from PyAstronomy import funcFit as fuf
        import matplotlib.pylab as plt
        import numpy as np

        # Create an instance of TraceAnalysis
        # telling it which file to use
        ta = fuf.TraceAnalysis("mcmcTA.emcee")

        # Use the burn-in from the previous example
        ta.setBurn(500)

        # Find sets of parameters
        # First, the lowest deviance set
        lds, index = ta.parameterSet(prescription="lowestDev")
        print("Lowest deviance set: ", lds)
        print("  at chain index: ", index)
        means = ta.parameterSet(prescription="mean")
        print("Set of mean values: ", means)
        medians = ta.parameterSet(prescription="median")
        print("Set of median values: ", means)

        # Create Voigt model and plot the models belonging
        # to the lowest deviance, mean, and median parameter
        # set.
        vp = fuf.Voigt1d()
        # Generate the model wavelength axis
        x = np.linspace(5490.0, 5510.0, 200)
        # Calculate and plot the models
        vp.assignValues(lds)
        plt.plot(x, vp.evaluate(x), "b.-")
        vp.assignValues(means)
        plt.plot(x, vp.evaluate(x), "r.-")
        vp.assignValues(medians)
        plt.plot(x, vp.evaluate(x), "g.-")
        #     plt.show()

        # Sixth example

        from PyAstronomy import funcFit as fuf

        # Create an instance of TraceAnalysis
        # telling it which file to use
        ta = fuf.TraceAnalysis("mcmcTA.emcee")

        # Use the burn-in from the previous example
        ta.setBurn(500)

        # Investigate a trace
        ta.plotTrace("mu")
        #    ta.show()
        # and its distribution.
        ta.plotHist("mu")
        #    ta.show()
        # Combine trace and distribution
        ta.plotTraceHist("mu")
        #    ta.show()
        # Plot correlations
        ta.plotCorr(parsList=["mu", "ad", "al"])

    #    ta.show()

    # =============================================================================
    # def testsanity_MCMCautoParameters(self):
    #   """
    #     Checking sanity of MCMCautoParameters
    #   """
    #   from PyAstronomy import funcFit as fuf
    #   import numpy as np
    #   import matplotlib.pylab as plt
    #
    #   x = np.linspace(0,30,1000)
    #   gauss = fuf.GaussFit1d()
    #   gauss["A"] = 1
    #   gauss["mu"] = 23.
    #   gauss["sig"] = 0.5
    #   yerr = np.random.normal(0., 0.05, len(x))
    #   y = gauss.evaluate(x) + yerr
    #   # This step is not necessary if <picky>=False in MCMCautoParameters.
    #   gauss.thaw(["A","mu","sig"])
    #   X0, lims, steps = gauss.MCMCautoParameters({"A":[0,10],"mu":3, "sig":[0.1,1.0]})
    #   gauss.fitMCMC(x, y, X0, lims, steps, yerr=yerr, iter=1000)
    #
    # =============================================================================
    #     plt.plot(x, y, 'k+')
    #     plt.plot(x, gauss.evaluate(x), 'r--')
    #     plt.show()

    def testsanity_EMCEEfirstexample(self):
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf

        # Before we can start fitting, we need something to fit.
        # So let us create some data...

        # Choose some signal-to-noise ratio
        snr = 25.0

        # Creating a Gaussian with some noise
        # Choose some parameters...
        gf = fuf.GaussFit1d()
        gf.assignValues(
            {"A": -5.0, "sig": 2.5, "mu": 10.0, "off": 1.0, "lin": 0.0})
        # Calculate profile
        x = arange(100) - 50.0
        y = gf.evaluate(x)
        # Add some noise
        y += random.normal(0.0, 1.0 / snr, x.size)

        # Define the free parameters
        gf.thaw(["A", "sig", "mu", "off"])

        # Start a fit (quite dispensable here)
        gf.fit(x, y, yerr=ones(x.size) / snr)

        # Say, we want 200 burn-in iterations and, thereafter,
        # 1000 further iterations (per walker).
        sampleArgs = {"iters": 1000, "burn": 200}

        # Start the sampling (ps could be used to continueb the sampling)
        ps = gf.fitEMCEE(x, y, yerr=ones(x.size) / snr, sampleArgs=sampleArgs)

        # Plot the distributions of the chains
        # NOTE: the order of the parameters in the chain object is the same
        #       as the order of the parameters returned by freeParamNames()
        for i, p in enumerate(gf.freeParamNames()):
            plt.subplot(len(gf.freeParamNames()), 1, i + 1)
            plt.hist(gf.emceeSampler.flatchain[::, i], label=p)
            plt.legend()

    #     plt.show()

    def testsanity_EMCEEpriorexample(self):
        # Import numpy and matplotlib
        from numpy import arange, sqrt, exp, pi, random, ones
        import matplotlib.pylab as plt

        # ... and now the funcFit package
        from PyAstronomy import funcFit as fuf
        import numpy as np

        # Before we can start fitting, we need something to fit.
        # So let us create some data...

        # Choose some signal-to-noise ratio
        snr = 25.0

        # Choosing an arbitrary constant and ...
        c = 10.0
        # ... an equally arbitrary number of data points
        npoint = 10

        # Define 'data'
        x = arange(npoint)
        y = np.ones(len(x)) * c
        # Add some noise
        y += random.normal(0.0, 1.0 / snr, x.size)

        # A funcFit object representing a constant
        pf = fuf.PolyFit1d(0)
        pf["c0"] = c

        # The only parameter shall be free
        pf.thaw("c0")

        # Say, we want 200 burn-in iterations and, thereafter,
        # 2500 further iterations (per walker).
        sampleArgs = {"iters": 2500, "burn": 200}

        # Start the sampling (ps could be used to continue the sampling)
        ps = pf.fitEMCEE(x, y, yerr=ones(x.size) / snr, sampleArgs=sampleArgs)
        print()

        # Plot the distributions of the chains
        # NOTE: the order of the parameters in the chain object is the same
        #       as the order of the parameters returned by freeParamNames()
        h = plt.hist(
            pf.emceeSampler.flatchain[::, 0], label="c0", density=True)
        # Construct "data points" in the middle of the bins
        xhist = (h[1][1:] + h[1][0:-1]) / 2.0
        yhist = h[0]

        # Fit the histogram using a Gaussian
        gf = fuf.GaussFit1d()
        gf.assignValues(
            {"A": 1.0, "mu": c, "sig": 1.0 / snr / np.sqrt(npoint)})
        # First fitting only "mu" is simply quite stable
        gf.thaw("mu")
        gf.fit(xhist, yhist)
        gf.thaw(["A", "sig"])
        gf.fit(xhist, yhist)

        print()
        print("  --- Sampling results ---")
        print(
            "Posterior estimate of constant: ",
            np.mean(pf.emceeSampler.flatchain[::, 0]),
        )
        print("Nominal error of the mean: ", 1.0 / snr / np.sqrt(npoint))
        print(
            "Estimate from Markov chain: ",
            np.std(pf.emceeSampler.flatchain[::, 0]),
            end=" ",
        )
        print(" and from Gaussian fit to distribution: ", gf["sig"])

        # Evaluate best-fit model ...
        xmodel = np.linspace(c - 10.0 / snr, c + 10.0 / snr, 250)
        ymodel = gf.evaluate(xmodel)
        # ... and plot
        plt.plot(xhist, yhist, "rp")
        plt.plot(xmodel, ymodel, "r--")
        plt.legend()
        #     plt.show()

        # Defining a prior on c0. Prior knowledge tells us that its value
        # is around 7. Let us choose the standard deviation of the prior so
        # that the estimate will lie in the middle between 7 and 10. Here we
        # exploit symmetry and make the prior information as strong as the
        # information contained in the likelihood function.
        priors = {
            "c0": fuf.FuFPrior("gaussian", sig=1.0 / snr / np.sqrt(npoint), mu=7.0)
        }

        # Start the sampling (ps could be used to continue the sampling)
        ps = pf.fitEMCEE(
            x, y, yerr=ones(x.size) / snr, sampleArgs=sampleArgs, priors=priors
        )

        print()
        print("  --- Sampling results with strong prior information ---")
        print(
            "Posterior estimate of constant: ",
            np.mean(pf.emceeSampler.flatchain[::, 0]),
            end=" ",
        )
        print(" +/-", np.std(pf.emceeSampler.flatchain[::, 0]))

        plt.hist(pf.emceeSampler.flatchain[::, 0], label="c0", density=True)

    #     plt.show()

    def testsanity_InstatiatePrior(self):
        from PyAstronomy import funcFit as fuf

        # Instantiate prior
        gp = fuf.FuFPrior("gaussian", sig=0.1, mu=1.0)

        # Current values (arbitrary)
        cvals = {"a": 1.4, "b": 0.86, "c": 1.1}

        # Get log(prior) for parameter "b"
        print(gp(cvals, "b"))

    def testsanity_sampleEMCEE_sampleFromGaussian(self):
        r"""
        Checking first sampleEMCEE example (sample from Gaussian distribution)
        """
        import numpy as np
        from PyAstronomy import funcFit as fuf
        import matplotlib.pylab as plt

        def lfGauss(v, sigma, mu):
            r"""
            Gaussian density

            Parameters
            ----------
            v : dictionary
                Holds current values of "x"
            mus, sigma : float
                Mean and standard deviation of the Gaussian. Specified via
                the `largs` argument.

            Returns
            -------
            lp : float
                Natural logarithm of the density.
            """
            result = 0.0
            # Log(density)
            result += -0.5 * np.log(2.0 * np.pi * sigma**2) - (v["x"] - mu) ** 2 / (
                2.0 * sigma**2
            )
            return result

        # Sampling arguments
        # burn: Number of burn-in steps per walker
        # iters: Number of iterations per walker
        sa = {"burn": 1000, "iters": 5000}

        # Starting values
        fv0 = {"x": 0.5}
        # Specify standard deviation and mean of Gaussian
        la = {"mu": 0.5, "sigma": 0.25}

        # Sample from distribution
        ps = fuf.sampleEMCEE(
            ["x"],
            fv0,
            lfGauss,
            largs=la,
            sampleArgs=sa,
            nwalker=4,
            dbfile="gauss.emcee",
        )
        print()

        # Use TraceAnalysis to look at chains
        ta = fuf.TraceAnalysis("gauss.emcee")
        print("Available chains: ", ta.availableParameters())

        print("Mean and STD of chain: ", np.mean(ta["x"]), np.std(ta["x"]))

        # Check distribution of chain
        # Plot histogram of chain
        #     plt.hist(ta["x"], 60, normed=True)
        # Overplot Gaussian model
        xx = np.linspace(la["mu"] - 6 * la["sigma"],
                         la["mu"] + 6 * la["sigma"], 1000)
        yy = (
            1.0
            / np.sqrt(2.0 * np.pi * la["sigma"] ** 2)
            * np.exp(-((xx - la["mu"]) ** 2) / (2.0 * la["sigma"] ** 2))
        )

    #     plt.plot(xx, yy, 'r--')
    #     plt.show()

    def testsanity_sampleEMCEE_estimateMuSig(self):
        r"""
        Checking sampleEMCEE example (estimate mu and sigma)
        """
        import numpy as np
        from PyAstronomy import funcFit as fuf

        def lfGaussMS(v, x=None):
            r"""
            Gaussian posterior with 1/sigma prior on sigma.

            Parameters
            ----------
            v : dictionary
                Holds current values of "sigma" and "mu"
            x : array
                The 'data' observed. Will be specified by the `largs` keyword.

            Returns
            -------
            lp : float
                Natural logarithm of the density.
            """
            if v["sigma"] < 0.0:
                # Penalize negative standard deviations
                return -1e20 * abs(v["sigma"])

            result = 0.0
            # Apply prior on sigma
            result -= np.log(v["sigma"])
            # Add log(likelihood)
            result += np.sum(
                -0.5 * np.log(2.0 * np.pi * v["sigma"] ** 2)
                - (x - v["mu"]) ** 2 / (2.0 * v["sigma"] ** 2)
            )
            return result

        # Sampling arguments
        # burn: Number of burn-in steps per walker
        # iters: Number of iterations per walker
        sa = {"burn": 1000, "iters": 5000}

        # Starting values
        fv0 = {"sigma": 1.0, "mu": 1.0}
        # 'Observed' data
        la = {"x": np.random.normal(0.0, 1.0, 1000)}

        print("Mean of 'data': ", np.mean(la["x"]))
        print("Standard deviation of 'data': ", np.std(la["x"]))

        # Scale width for distributing the walkers
        s = {"mu": 0.01, "sigma": 0.5}
        ps = fuf.sampleEMCEE(
            ["mu", "sigma"],
            fv0,
            lfGaussMS,
            largs=la,
            sampleArgs=sa,
            nwalker=4,
            scales=s,
            dbfile="musig.emcee",
        )
        print()

        # Use TraceAnalysis to look at chains
        ta = fuf.TraceAnalysis("musig.emcee")
        print("Available chains: ", ta.availableParameters())


#     ta.plotTraceHist('mu')
#     ta.show()
#
#     ta.plotTraceHist('sigma')
#     ta.show()



from .importCheck import ImportCheck, pyaimportallfrom

_ic = ImportCheck(["numpy", "scipy", "ssl"], required=["numpy", "scipy"])

from .sanity import *
from .nestedLoop import NestedLoop
from .invertIndexSelection import invertIndexSelection
from .fuzzyMatch import fuzzyMatch
from .output import matrix2doutput
from .simiof import SimIOF

from .mtools import *


from __future__ import division
import numpy as np

def applyProperMotion(ra, dec, pmra, pmdec, dt, fixes=1):
    """
    Apply proper motion to coordinates
    
    Parameters
    ----------
    ra, dec : float or array
        Coordinates at reference epoch
    pmra, pmdec : float or array
        Proper motion in mas/yr.
    dt : float
        Elapsed time [yr]
    fixes : int, optional
        Number of steps in which the proper motion is applied.
        Default is one.
    """
    
    # mas/yr -> deg/yr
    pmra = pmra / 3.6e6
    pmdec = pmdec / 3.6e6
    
    dtt = dt / fixes
    for i in range(fixes):
        decnew = dec + pmdec*dtt
        ra = ra + pmra*dtt / np.cos(np.deg2rad((dec+decnew)/2))
        dec = decnew
        
    return ra, dec
    


from scipy.stats import f
import numpy
from scipy.special import erf
from scipy.optimize import fsolve
from PyAstronomy.pyaC import pyaErrors as PE


def chi2test():
    pass


def bayesFactors(model1, model2):
    raise (PE.PyAUnclassifiedError(
        "The function 'bayesFactors' has moved to funcFit.utils."))


def ftest(chi1, chi2, dof1, dof2, compute_ratio_of_variance=False):
    r"""
    Performs an F-test. 

    Parameters
    ----------
    chi1 : float
        The chi-squared statistic or variance for model 1.
    chi2 : float
        The chi-squared statistic or variance for model 2.
        Should be better (lower) than `chi1`.
    dof1 : integer
        The number of degrees of freedom of model 1
    dof2 : integer
        The number of degrees of freedom of model 2.
        Should be lower than `dof1`.
    compute_ratio_of_variance : boolean, optional
        Determines how the F-statistics is computed.
        If False (default), the "ANOVA F-test in regression analysis
        of nested nonlinear models" is applied.
        If True, "Fisher's classical test for the equality
        of two variances" is used.

    Returns
    -------
    Test results : dictionary
        - "F statistic" - Value of F statistic.
        - "p-value" - Probability to obtain an F-value as large as
          or larger than observed assuming H_0 to be true
        - "Gaussian sigma level" - Same probability converted to
          Gaussian standard deviations ("sigmas")


    Notes
    -----

    The "method of least squares" is widely used in parameter estimation.
    Much of its appeal lies in its simplicity: Minimizing the sum of squared
    differences between the model and the data.

    The backbone of least squares is the classical multiple regression
    analysis using linear models to relate several independent variables to
    a dependent variable. However, many physical models are nonlinear;
    this is the realm of nonlinear least squares regression methods.

    The aim of model fitting is to explain as much of the variation in the
    dependent variable as possible from information contained in the
    independent variables. The contributions of the independent variables
    to the model are measured by partitions of the total sum of squares of
    the difference of data and model ("analysis of variance", ANOVA).

    **ANOVA F-test in regression analysis for nested nonlinear models**

    The sum of squares for any hypothesis can be determined from the
    difference between the residual sums of squares (:math:`RSS`) of two models: The
    so-called "full" model, in the context of which the null hypothesis is to be
    tested, and the "reduced" model, which is derived from
    the full model by imposing additional constraints specified by the null hypothesis;
    setting one parameter to zero is a common example.
    The reduced model is a special case of the full model and, hence, its
    residual sum of squares must not be lower than the residual
    sum of squares for the full model.

    Accounting for :math:`k` independent constraints on the full model and :math:`N` data
    points, we compute

    .. math:: Q = RSS(reduced) - RSS(full)

    which has :math:`k` degrees of freedom. Here, :math:`k` corresponds to the
    difference in the number of parameters---or equivalently, the difference
    in the degrees of freedoms---between the full and reduced model.
    The :math:`F`-test of the null hypothesis can
    then be computed as

    .. math:: \hat{F} = (Q/k) / s^2

    where :math:`s^2` is an unbiased estimate of :math:`\sigma^2`, e.g., derived from the
    full model. In the case of error-weighted data, the :math:`F` statistic reads

    .. math:: \hat{F} = \frac{(\chi^2_{reduced}-\chi^2_{full})/(\nu_{reduced}-\nu_{full})}{\chi^2_{full}/\nu_{full}}

    where :math:`\nu` denotes the number of degrees of freedom, which may be calculated as
    :math:`N-p-1` given a model with :math:`p` parameters and an additional constant.
    The expectation value of :math:`F` is :math:`1`, thus, if :math:`\hat{F}=1` 
    there is no significant difference
    between the RSS of both models. If, however, :math:`F` deviates from :math:`1`, we can compute
    the probability for :math:`F` to equal or exceed the obtained value :math:`\hat{F}` by

    .. math:: Prob(F\geq \hat{F}) = CDF(F( \nu_{reduced}-\nu_{full}, \nu_{full} ))

    For details on the principles of the formalism,
    see Sect. 4.5.4 in Rawlings' "Applied Regression Analysis".

    In case of nonlinear models, the distribution of the least-square
    parameter estimates only approximately follows the normal distribution.
    Then, the :math:`F` statistic (in this case called "Wald statistic") also
    holds approximately.
    The validity of the method depends on how well the model is represented
    by a linear approximation in the parameters ("parameter effects curvature").
    For details, we refer to Sect. 15.3 in Rawlings' "Applied Regression Analysis"
    and references therein.

    **Fisher's classical test for the equality of two variances**

    If the two vying models cannot be treated as *nested*, in the sense that
    the full model encompasses the reduced model, we can directly compare
    the model variances with an F-test.

    Let :math:`s_1` and :math:`s_2` denote unbiased estimates of the variances
    of two independent, normally distributed populations of size :math:`N_1` and :math:`N_2`,
    respectively. Under the null hypothesis
    both variances are equal:

    .. math:: H_0: s_1^2=s_2^2, \hspace{20pt} H_1: s_1^2>s_2^2

    Then the quantity

    .. math:: \hat{F} = \frac{s_1^2}{s_2^2}

    is :math:`F`-distributed with :math:`N_1-1` and :math:`N_2-1` degrees of freedom.
    In the case of weighted errors we have

    .. math:: \hat{F} = \frac{\chi_1^2/\nu_1}{\chi_2^2/\nu_2}

    and :math:`\hat{F}` is distributed according to an
    :math:`F`-distribution with :math:`\nu_1` and :math:`\nu_2`
    degrees of freedom.

    **References**

    John O. Rawlings, Sastry G. Pantula, David A. Dickey
    *Applied Regression Analysis: A Research Tool*, Second Edition
    Springer, New York, NY, 1998

    """
    # Some consistency checks
    if not dof2 < dof1:
        raise (PE.PyAValError("The number of degrees of freedom in the new fit, dof2, should be less" +
                              " than the number of degrees of freedom in the original fit, dof1."))
    if not chi2 < chi1:
        raise (PE.PyAValError(
            "The new fit (2) should have lower chi-squared than the original fit (1)."))

    # Compute the test statistic and the probability
    if compute_ratio_of_variance:
        F = (chi1 / dof1) / (chi2 / dof2)
        prob = f.cdf(F, dof1, dof2)
    else:
        F = ((chi1 - chi2) / (dof1 - dof2)) / (chi2 / dof2)
        prob = f.cdf(F, dof1 - dof2, dof2)

    # Convert probability into Gaussian 'sigmas'
    def err(n, prob):
        return erf(n / numpy.sqrt(2)) - prob
    sigmalvl = float(fsolve(err, 1.0, args=(prob,)))

    return {"F statistic": F, "p-value": 1. - prob, "Gaussian sigma level": sigmalvl}
